import socket

from queue import Queue
from unittest import mock

import confluent_kafka
import pytest

from gevent.server import StreamServer
from prometheus_client import REGISTRY

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate.frameworks.queue_consumer.kafka import FastConsumerFactory
from baseplate.frameworks.queue_consumer.kafka import InOrderConsumerFactory
from baseplate.frameworks.queue_consumer.kafka import KAFKA_ACTIVE_MESSAGES
from baseplate.frameworks.queue_consumer.kafka import KAFKA_PROCESSED_TOTAL
from baseplate.frameworks.queue_consumer.kafka import KAFKA_PROCESSING_TIME
from baseplate.frameworks.queue_consumer.kafka import KafkaConsumerPrometheusLabels
from baseplate.frameworks.queue_consumer.kafka import KafkaConsumerWorker
from baseplate.frameworks.queue_consumer.kafka import KafkaMessageHandler
from baseplate.lib import metrics


@pytest.fixture
def context():
    context = mock.Mock(spec=RequestContext)
    context.metrics = mock.Mock(spec=metrics.Client)
    return context


@pytest.fixture
def span():
    sp = mock.MagicMock(spec=ServerSpan)
    sp.make_child().__enter__.return_value = mock.MagicMock()
    return sp


@pytest.fixture
def baseplate(context, span):
    bp = mock.MagicMock(spec=Baseplate)
    bp._metrics_client = None
    bp.make_server_span().__enter__.return_value = span
    bp.make_context_object.return_value = context
    # Reset the mock calls since setting up the span actually triggers a "call"
    # to bp.make_server_span
    bp.reset_mock()
    return bp


@pytest.fixture
def name():
    return "kafka_consumer.test_group"


class TestKafkaMessageHandler:
    def setup(self):
        KAFKA_PROCESSING_TIME.clear()
        KAFKA_PROCESSED_TOTAL.clear()
        KAFKA_ACTIVE_MESSAGES.clear()

    @pytest.fixture
    def message(self):
        msg = mock.Mock(spec=confluent_kafka.Message)
        msg.topic.return_value = "topic_1"
        msg.key.return_value = "key_1"
        msg.partition.return_value = 3
        msg.offset.return_value = 33
        msg.timestamp.return_value = 123456
        msg.value.return_value = b"message-payload"
        msg.error.return_value = None
        return msg

    @mock.patch("baseplate.frameworks.queue_consumer.kafka.time")
    @pytest.mark.parametrize("prometheus_client_name", [None, "my_kafka_client_name"])
    def test_handle(self, time, context, span, baseplate, name, message, prometheus_client_name):
        time.time.return_value = 2.0
        time.perf_counter.side_effect = [1, 2]

        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name=prometheus_client_name if prometheus_client_name is not None else "",
            kafka_topic="topic_1",
        )

        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock(
            return_value={"endpoint_timestamp": 1000.0, "body": "some text"}
        )
        on_success_fn = mock.Mock()

        mock_timer = mock.Mock()
        context.metrics.timer.return_value = mock_timer

        mock_gauge = mock.Mock()
        context.metrics.gauge.return_value = mock_gauge

        if prometheus_client_name is None:
            handler = KafkaMessageHandler(
                baseplate, name, handler_fn, message_unpack_fn, on_success_fn
            )
        else:
            handler = KafkaMessageHandler(
                baseplate,
                name,
                handler_fn,
                message_unpack_fn,
                on_success_fn,
                prometheus_client_name=prometheus_client_name,
            )
        handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, f"{name}.handler")
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("kafka.topic", "topic_1"),
                mock.call("kafka.key", "key_1"),
                mock.call("kafka.partition", 3),
                mock.call("kafka.offset", 33),
                mock.call("kafka.timestamp", 123456),
            ],
            any_order=True,
        )
        message_unpack_fn.assert_called_once_with(b"message-payload")
        handler_fn.assert_called_once_with(
            context, {"endpoint_timestamp": 1000.0, "body": "some text"}, message
        )
        on_success_fn.assert_called_once_with(
            context, {"endpoint_timestamp": 1000.0, "body": "some text"}, message
        )

        context.metrics.timer.assert_called_once_with(f"{name}.topic_1.latency")
        mock_timer.send.assert_called_once_with(1.0)

        context.metrics.gauge.assert_called_once_with(f"{name}.topic_1.offset.3")
        mock_gauge.replace.assert_called_once_with(33)

        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"kafka_success": "true", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"kafka_success": "true"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{KAFKA_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )

    def test_handle_no_endpoint_timestamp(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock(return_value={"body": "some text"})
        on_success_fn = mock.Mock()

        mock_gauge = mock.Mock()
        context.metrics.gauge.return_value = mock_gauge

        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name="",
            kafka_topic="topic_1",
        )

        handler = KafkaMessageHandler(baseplate, name, handler_fn, message_unpack_fn, on_success_fn)
        handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, f"{name}.handler")
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("kafka.topic", "topic_1"),
                mock.call("kafka.key", "key_1"),
                mock.call("kafka.partition", 3),
                mock.call("kafka.offset", 33),
                mock.call("kafka.timestamp", 123456),
            ],
            any_order=True,
        )
        message_unpack_fn.assert_called_once_with(b"message-payload")
        handler_fn.assert_called_once_with(context, {"body": "some text"}, message)
        on_success_fn.assert_called_once_with(context, {"body": "some text"}, message)

        context.metrics.timer.assert_not_called()

        context.metrics.gauge.assert_called_once_with(f"{name}.topic_1.offset.3")
        mock_gauge.replace.assert_called_once_with(33)

        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"kafka_success": "true", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"kafka_success": "true"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{KAFKA_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )

    def test_handle_kafka_error(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock()
        on_success_fn = mock.Mock()

        # we can't actually create an instance of KafkaError, so use a mock
        error_mock = mock.Mock()
        error_mock.str.return_value = "kafka error"
        message.error.return_value = error_mock

        handler = KafkaMessageHandler(baseplate, name, handler_fn, message_unpack_fn, on_success_fn)

        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name="",
            kafka_topic="topic_1",
        )

        with pytest.raises(ValueError):
            handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, f"{name}.handler")
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_not_called()
        message_unpack_fn.assert_not_called()
        handler_fn.assert_not_called()
        on_success_fn.assert_not_called()
        context.metrics.timer.assert_not_called()
        context.metrics.gauge.assert_not_called()

        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"kafka_success": "false", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"kafka_success": "false"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{KAFKA_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )

    def test_handle_unpack_error(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock(side_effect=ValueError("something bad happened"))
        on_success_fn = mock.Mock()

        context.span = span

        handler = KafkaMessageHandler(baseplate, name, handler_fn, message_unpack_fn, on_success_fn)

        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name="",
            kafka_topic="topic_1",
        )

        handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, f"{name}.handler")
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("kafka.topic", "topic_1"),
                mock.call("kafka.key", "key_1"),
                mock.call("kafka.partition", 3),
                mock.call("kafka.offset", 33),
                mock.call("kafka.timestamp", 123456),
            ],
            any_order=True,
        )
        span.incr_tag.assert_called_once_with(f"{name}.topic_1.invalid_message")
        message_unpack_fn.assert_called_once_with(b"message-payload")
        handler_fn.assert_not_called()
        on_success_fn.assert_not_called()
        context.metrics.timer.assert_not_called()
        context.metrics.gauge.assert_not_called()

        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"kafka_success": "false", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"kafka_success": "false"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{KAFKA_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )

    def test_handle_handler_error(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock(side_effect=ValueError("something went wrong"))
        message_unpack_fn = mock.Mock(
            return_value={"endpoint_timestamp": 1000.0, "body": "some text"}
        )
        on_success_fn = mock.Mock()

        handler = KafkaMessageHandler(baseplate, name, handler_fn, message_unpack_fn, on_success_fn)

        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name="",
            kafka_topic="topic_1",
        )

        with pytest.raises(ValueError):
            handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, f"{name}.handler")
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("kafka.topic", "topic_1"),
                mock.call("kafka.key", "key_1"),
                mock.call("kafka.partition", 3),
                mock.call("kafka.offset", 33),
                mock.call("kafka.timestamp", 123456),
            ],
            any_order=True,
        )
        message_unpack_fn.assert_called_once_with(b"message-payload")
        handler_fn.assert_called_once_with(
            context, {"endpoint_timestamp": 1000.0, "body": "some text"}, message
        )
        on_success_fn.assert_not_called()
        context.metrics.timer.assert_not_called()
        context.metrics.gauge.assert_not_called()

        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"kafka_success": "false", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{KAFKA_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"kafka_success": "false"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{KAFKA_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )


@pytest.fixture
def bootstrap_servers():
    return "127.0.0.1:9092"


@pytest.fixture
def group_id():
    return "test_service.test_group"


@pytest.fixture
def topics():
    return ["topic_1"]


class TestInOrderConsumerFactory:
    @mock.patch("confluent_kafka.Consumer")
    def test_make_kafka_consumer(
        self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics
    ):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer
        extra_config = {"queued.max.messages.kbytes": 10000}

        _consumer = InOrderConsumerFactory.make_kafka_consumer(
            bootstrap_servers, group_id, topics, extra_config
        )

        assert _consumer == mock_consumer

        kafka_consumer.assert_called_once_with(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": group_id,
                "auto.offset.reset": "latest",
                "heartbeat.interval.ms": 3000,
                "session.timeout.ms": 10000,
                "max.poll.interval.ms": 300000,
                "enable.auto.commit": "false",
                "queued.max.messages.kbytes": 10000,
            }
        )
        mock_consumer.subscribe.assert_called_once()
        assert mock_consumer.subscribe.call_args_list[0][0][0] == topics

    @mock.patch("confluent_kafka.Consumer")
    def test_make_kafka_consumer_unknown_topic(
        self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics
    ):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer
        extra_config = {"queued.max.messages.kbytes": 100}

        with pytest.raises(AssertionError):
            InOrderConsumerFactory.make_kafka_consumer(
                bootstrap_servers, group_id, ["topic_4"], extra_config
            )

        kafka_consumer.assert_called_once_with(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": group_id,
                "auto.offset.reset": "latest",
                "heartbeat.interval.ms": 3000,
                "session.timeout.ms": 10000,
                "max.poll.interval.ms": 300000,
                "enable.auto.commit": "false",
                "queued.max.messages.kbytes": 100,
            }
        )
        mock_consumer.subscribe.assert_not_called()

    @mock.patch("confluent_kafka.Consumer")
    def test_init(self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer

        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock()
        health_check_fn = mock.Mock()
        factory = InOrderConsumerFactory.new(
            name=name,
            baseplate=baseplate,
            bootstrap_servers=bootstrap_servers,
            group_id=group_id,
            topics=topics,
            handler_fn=handler_fn,
            message_unpack_fn=message_unpack_fn,
            health_check_fn=health_check_fn,
        )
        assert factory.name == name
        assert factory.baseplate == baseplate
        assert factory.handler_fn == handler_fn
        assert factory.message_unpack_fn == message_unpack_fn
        assert factory.health_check_fn == health_check_fn
        assert factory.consumer == mock_consumer

    @pytest.fixture
    def make_queue_consumer_factory(self, name, baseplate, bootstrap_servers, group_id, topics):
        @mock.patch("confluent_kafka.Consumer")
        def _make_queue_consumer_factory(kafka_consumer, health_check_fn=None):
            mock_consumer = mock.Mock()
            mock_consumer.list_topics.return_value = mock.Mock(
                topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
            )
            kafka_consumer.return_value = mock_consumer

            return InOrderConsumerFactory.new(
                name=name,
                baseplate=baseplate,
                bootstrap_servers=bootstrap_servers,
                group_id=group_id,
                topics=topics,
                handler_fn=lambda ctx, data, msg: True,
                message_unpack_fn=lambda b: {},
                health_check_fn=health_check_fn,
            )

        return _make_queue_consumer_factory

    def test_build_pump_worker(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        work_queue = Queue(maxsize=10)
        pump = factory.build_pump_worker(work_queue)
        assert isinstance(pump, KafkaConsumerWorker)
        assert pump.consumer == factory.consumer
        assert pump.work_queue == work_queue

    def test_build_message_handler(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        handler = factory.build_message_handler()
        assert isinstance(handler, KafkaMessageHandler)
        assert handler.baseplate == factory.baseplate
        assert handler.name == factory.name
        assert handler.handler_fn == factory.handler_fn
        assert handler.message_unpack_fn == factory.message_unpack_fn
        assert handler.on_success_fn.__name__ == "commit_offset"

    def test_build_multiple_message_handlers(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()

        factory.build_message_handler()

        with pytest.raises(AssertionError):
            factory.build_message_handler()

    @pytest.mark.parametrize("health_check_fn", [None, lambda req: True])
    def test_build_health_checker(self, health_check_fn, make_queue_consumer_factory):
        factory = make_queue_consumer_factory(health_check_fn=health_check_fn)
        listener = mock.Mock(spec=socket.socket)
        health_checker = factory.build_health_checker(listener)
        assert isinstance(health_checker, StreamServer)


class TestFastConsumerFactory:
    @mock.patch("confluent_kafka.Consumer")
    def test_make_kafka_consumer(
        self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics
    ):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer

        _consumer = FastConsumerFactory.make_kafka_consumer(bootstrap_servers, group_id, topics)

        assert _consumer == mock_consumer

        kafka_consumer.assert_called_once_with(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": group_id,
                "auto.offset.reset": "latest",
                "heartbeat.interval.ms": 3000,
                "session.timeout.ms": 10000,
                "max.poll.interval.ms": 300000,
                "enable.auto.commit": "true",
                "auto.commit.interval.ms": 5000,
                "enable.auto.offset.store": "true",
                "on_commit": FastConsumerFactory._commit_callback,
            }
        )
        mock_consumer.subscribe.assert_called_once()
        assert mock_consumer.subscribe.call_args_list[0][0][0] == topics

    @mock.patch("confluent_kafka.Consumer")
    def test_make_kafka_consumer_unknown_topic(
        self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics
    ):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer

        with pytest.raises(AssertionError):
            FastConsumerFactory.make_kafka_consumer(bootstrap_servers, group_id, ["topic_4"])

        kafka_consumer.assert_called_once_with(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": group_id,
                "auto.offset.reset": "latest",
                "heartbeat.interval.ms": 3000,
                "session.timeout.ms": 10000,
                "max.poll.interval.ms": 300000,
                "enable.auto.commit": "true",
                "auto.commit.interval.ms": 5000,
                "enable.auto.offset.store": "true",
                "on_commit": FastConsumerFactory._commit_callback,
            }
        )
        mock_consumer.subscribe.assert_not_called()

    @mock.patch("confluent_kafka.Consumer")
    def test_init(self, kafka_consumer, name, baseplate, bootstrap_servers, group_id, topics):
        mock_consumer = mock.Mock()
        mock_consumer.list_topics.return_value = mock.Mock(
            topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
        )
        kafka_consumer.return_value = mock_consumer

        handler_fn = mock.Mock()
        message_unpack_fn = mock.Mock()
        health_check_fn = mock.Mock()
        factory = FastConsumerFactory.new(
            name=name,
            baseplate=baseplate,
            bootstrap_servers=bootstrap_servers,
            group_id=group_id,
            topics=topics,
            handler_fn=handler_fn,
            message_unpack_fn=message_unpack_fn,
            health_check_fn=health_check_fn,
        )
        assert factory.name == name
        assert factory.baseplate == baseplate
        assert factory.handler_fn == handler_fn
        assert factory.message_unpack_fn == message_unpack_fn
        assert factory.health_check_fn == health_check_fn
        assert factory.consumer == mock_consumer

    @pytest.fixture
    def make_queue_consumer_factory(self, name, baseplate, bootstrap_servers, group_id, topics):
        @mock.patch("confluent_kafka.Consumer")
        def _make_queue_consumer_factory(kafka_consumer, health_check_fn=None):
            mock_consumer = mock.Mock()
            mock_consumer.list_topics.return_value = mock.Mock(
                topics={"topic_1": mock.Mock(), "topic_2": mock.Mock(), "topic_3": mock.Mock()}
            )
            kafka_consumer.return_value = mock_consumer

            return FastConsumerFactory.new(
                name=name,
                baseplate=baseplate,
                bootstrap_servers=bootstrap_servers,
                group_id=group_id,
                topics=topics,
                handler_fn=lambda ctx, data, msg: True,
                message_unpack_fn=lambda b: {},
                health_check_fn=health_check_fn,
            )

        return _make_queue_consumer_factory

    def test_build_pump_worker(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        work_queue = Queue(maxsize=10)
        pump = factory.build_pump_worker(work_queue)
        assert isinstance(pump, KafkaConsumerWorker)
        assert pump.consumer == factory.consumer
        assert pump.work_queue == work_queue

    def test_build_message_handler(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        handler = factory.build_message_handler()
        assert isinstance(handler, KafkaMessageHandler)
        assert handler.baseplate == factory.baseplate
        assert handler.name == factory.name
        assert handler.handler_fn == factory.handler_fn
        assert handler.message_unpack_fn == factory.message_unpack_fn
        assert handler.on_success_fn is None

    @pytest.mark.parametrize("health_check_fn", [None, lambda req: True])
    def test_build_health_checker(self, health_check_fn, make_queue_consumer_factory):
        factory = make_queue_consumer_factory(health_check_fn=health_check_fn)
        listener = mock.Mock(spec=socket.socket)
        health_checker = factory.build_health_checker(listener)
        assert isinstance(health_checker, StreamServer)


class TestKafkaConsumerWorker:
    @pytest.fixture
    def consumer_worker(self, baseplate, name):
        mock_consumer = mock.Mock()
        mock_queue = mock.Mock(spec=Queue)
        return KafkaConsumerWorker(baseplate, name, mock_consumer, mock_queue)

    def test_initial_state(self, consumer_worker):
        assert consumer_worker.started is False
        assert consumer_worker.stopped is False

    @mock.patch("baseplate.frameworks.queue_consumer.kafka.time")
    def test_run(self, time, span, baseplate, name, consumer_worker):
        msg1 = mock.Mock()
        msg2 = mock.Mock()
        msg3 = mock.Mock()
        consumer_worker.consumer.consume.side_effect = [(), (msg1,), (msg2,), (msg3,)]

        with mock.patch.object(
            KafkaConsumerWorker, "stopped", create=True, new_callable=mock.PropertyMock
        ) as stopped_value:
            stopped_value.side_effect = [False, False, False, True]
            consumer_worker.run()

        baseplate.make_context_object.mock_calls == [mock.call(), mock.call(), mock.call()]
        baseplate.make_server_span.mock_calls == [
            mock.call(context, f"{name}.pump"),
            mock.call(context, f"{name}.pump"),
            mock.call(context, f"{name}.pump"),
        ]
        span.make_child.mock_calls == [
            mock.call("kafka.consume"),
            mock.call("kafka.work_queue_put"),
            mock.call("kafka.consume"),
            mock.call("kafka.work_queue_put"),
            mock.call("kafka.consume"),
            mock.call("kafka.work_queue_put"),
        ]

        assert consumer_worker.started is True
        assert consumer_worker.consumer.consume.mock_calls == [
            mock.call(num_messages=1, timeout=0),
            mock.call(num_messages=1, timeout=0),
            mock.call(num_messages=1, timeout=0),
        ]

        time.sleep.assert_called_once_with(1)
        assert consumer_worker.work_queue.put.mock_calls == [mock.call(msg1), mock.call(msg2)]

    def test_stop(self, consumer_worker):
        consumer_worker.stop()
        assert consumer_worker.stopped is True



import unittest

from unittest import mock

try:
    from pymemcache.exceptions import MemcacheClientError
except ImportError:
    raise unittest.SkipTest("pymemcache is not installed")

from baseplate.clients.memcache import MemcacheClient, MonitoredMemcacheConnection, make_keys_str
from baseplate import Baseplate, LocalSpan, ServerSpan

from . import TestBaseplateObserver, get_endpoint_or_skip_container


memcached_endpoint = get_endpoint_or_skip_container("memcached", 11211)


class MemcacheIntegrationTests(unittest.TestCase):
    def setUp(self):
        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate({"memcache.endpoint": str(memcached_endpoint)})
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context({"memcache": MemcacheClient()})

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple(self):
        with self.server_span:
            self.context.memcache.get("whatever")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "memcache.get")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_error(self):
        with self.server_span:
            with self.assertRaises(MemcacheClientError):
                self.context.memcache.cas("key", b"value", b"whatever")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "memcache.cas")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)


class MonitoredMemcacheConnectionIntegrationTests(unittest.TestCase):
    def setUp(self):
        self.mocked_pool = mock.Mock(server=memcached_endpoint.address)
        self.context_name = "memcache"
        self.server_span = mock.MagicMock(spec_set=ServerSpan)
        self.local_span = mock.MagicMock(spec_set=LocalSpan)
        self.local_span.__enter__.return_value = self.local_span
        self.server_span.make_child.return_value = self.local_span
        self.connection = MonitoredMemcacheConnection(
            self.context_name, self.server_span, self.mocked_pool
        )

        self.key = b"key"
        self.value = b"value"
        self.expire = 0
        self.noreply = False

    def test_close(self):
        self.connection.close()
        self.mocked_pool.close.assert_called_with()
        self.assertEqual(self.local_span.set_tag.call_count, 1)

    def test_set(self):
        self.connection.set(self.key, self.value, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.set.assert_called_with(
            self.key, self.value, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_replace(self):
        self.connection.replace(self.key, self.value, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.replace.assert_called_with(
            self.key, self.value, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_append(self):
        self.connection.append(self.key, self.value, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.append.assert_called_with(
            self.key, self.value, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_prepend(self):
        self.connection.prepend(self.key, self.value, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.prepend.assert_called_with(
            self.key, self.value, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_cas(self):
        cas = b"cascas"
        self.connection.cas(self.key, self.value, cas, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.cas.assert_called_with(
            self.key, self.value, cas, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 5)

    def test_get(self):
        self.connection.get(self.key)
        self.mocked_pool.get.assert_called_with(self.key)
        self.assertEqual(self.local_span.set_tag.call_count, 2)

    def test_gets(self):
        self.connection.gets(self.key)
        self.mocked_pool.gets.assert_called_with(self.key, default=None, cas_default=None)
        self.assertEqual(self.local_span.set_tag.call_count, 2)

    def test_gets_many(self):
        keys = [b"key1", b"key2"]
        self.connection.gets_many(keys)
        self.mocked_pool.gets_many.assert_called_with(keys)
        self.assertEqual(self.local_span.set_tag.call_count, 3)

    def test_delete(self):
        self.connection.delete(self.key, noreply=self.noreply)
        self.mocked_pool.delete.assert_called_with(self.key, noreply=self.noreply)
        self.assertEqual(self.local_span.set_tag.call_count, 3)

    def test_delete_many(self):
        keys = [b"key1", b"key2"]
        self.connection.delete_many(keys, noreply=self.noreply)
        self.mocked_pool.delete_many.assert_called_with(keys, noreply=self.noreply)
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_add(self):
        self.connection.add(self.key, self.value, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.add.assert_called_with(
            self.key, self.value, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_incr(self):
        value = 1
        self.connection.incr(self.key, value, noreply=self.noreply)
        self.mocked_pool.incr.assert_called_with(self.key, value, noreply=self.noreply)
        self.assertEqual(self.local_span.set_tag.call_count, 3)

    def test_decr(self):
        value = 1
        self.connection.decr(self.key, value, noreply=self.noreply)
        self.mocked_pool.decr.assert_called_with(self.key, value, noreply=self.noreply)
        self.assertEqual(self.local_span.set_tag.call_count, 3)

    def test_touch(self):
        self.connection.touch(self.key, expire=self.expire, noreply=self.noreply)
        self.mocked_pool.touch.assert_called_with(
            self.key, expire=self.expire, noreply=self.noreply
        )
        self.assertEqual(self.local_span.set_tag.call_count, 4)

    def test_flush_all(self):
        delay = 0
        self.connection.flush_all(delay=delay, noreply=self.noreply)
        self.mocked_pool.flush_all.assert_called_with(delay=delay, noreply=self.noreply)
        self.assertEqual(self.local_span.set_tag.call_count, 3)

    def test_quit(self):
        self.connection.quit()
        self.mocked_pool.quit.assert_called_with()
        self.assertEqual(self.local_span.set_tag.call_count, 1)


class MakeKeysStrTests(unittest.TestCase):
    def test_bytes(self):
        expected_string = "key_1,key_2"
        keys = [b"key_1", b"key_2"]
        self.assertEqual(expected_string, make_keys_str(keys))

    def test_str(self):
        expected_string = "key_1,key_2"
        keys = ["key_1", "key_2"]
        self.assertEqual(expected_string, make_keys_str(keys))



import socket
import unittest

from unittest import mock

from baseplate.lib import config
from baseplate.lib import metrics


EXAMPLE_ENDPOINT = config.EndpointConfiguration(socket.AF_INET, ("127.0.0.1", 1234))


class MetricJoinTests(unittest.TestCase):
    def test_single_node(self):
        joined = metrics._metric_join(b"single")
        self.assertEqual(joined, b"single")

    def test_two_nodes(self):
        joined = metrics._metric_join(b"first", b"second")
        self.assertEqual(joined, b"first.second")

    def test_subpaths(self):
        joined = metrics._metric_join(b"first.second", b"third.fourth")
        self.assertEqual(joined, b"first.second.third.fourth")

    def test_subpaths_with_trailing_dot(self):
        joined = metrics._metric_join(b"first.", b"second")
        self.assertEqual(joined, b"first.second")

    def test_join_null_nodes(self):
        joined = metrics._metric_join(b"first.", None, b"second", None)
        self.assertEqual(joined, b"first.second")


class FormatTagsTest(unittest.TestCase):
    def test_no_tags(self):
        formatted_tags = metrics._format_tags(None)
        self.assertIsNone(formatted_tags)

    def test_tags(self):
        tags = {"success": True, "error": False}
        formatted_tags = metrics._format_tags(tags)
        self.assertEqual(formatted_tags, b",success=True,error=False")


class NullTransportTests(unittest.TestCase):
    @mock.patch("socket.socket")
    def test_nothing_sent(self, mock_make_socket):
        transport = metrics.NullTransport(log_if_unconfigured=False)
        transport.send(b"metric")
        self.assertEqual(mock_make_socket.call_count, 0)


class RawTransportTests(unittest.TestCase):
    @mock.patch("socket.socket")
    def test_sent_immediately(self, mock_make_socket):
        mocket = mock_make_socket.return_value
        transport = metrics.RawTransport(EXAMPLE_ENDPOINT)

        self.assertEqual(mocket.connect.call_args, mock.call(("127.0.0.1", 1234)))

        transport.send(b"metric")

        self.assertEqual(mocket.sendall.call_count, 1)
        self.assertEqual(mocket.sendall.call_args, mock.call(b"metric"))


class BufferedTransportTests(unittest.TestCase):
    @mock.patch("socket.socket")
    def test_buffered(self, mock_make_socket):
        mocket = mock_make_socket.return_value
        raw_transport = metrics.RawTransport(EXAMPLE_ENDPOINT)
        transport = metrics.BufferedTransport(raw_transport)
        transport.send(b"a")
        transport.send(b"b")
        transport.send(b"c")
        self.assertEqual(mocket.sendall.call_count, 0)
        transport.flush()

        self.assertEqual(mocket.sendall.call_count, 1)
        self.assertEqual(mocket.sendall.call_args, mock.call(b"a\nb\nc"))

    def test_buffered_exception_is_caught(self):
        raw_transport = metrics.RawTransport(EXAMPLE_ENDPOINT)
        transport = metrics.BufferedTransport(raw_transport)
        transport.send(b"x" * 65536)

        with self.assertRaises(metrics.MessageTooBigTransportError):
            transport.flush()


class BaseClientTests(unittest.TestCase):
    def test_encode_namespace(self):
        transport = mock.Mock(spec=metrics.NullTransport)

        client = metrics.BaseClient(transport, "name")
        self.assertEqual(client.namespace, b"name")

        with self.assertRaises(UnicodeEncodeError):
            metrics.BaseClient(transport, "☃")


class BaseClientFactoriesTests(unittest.TestCase):
    def setUp(self):
        transport = mock.Mock(spec=metrics.NullTransport)
        self.client = metrics.BaseClient(transport, "namespace")

    def test_make_timer(self):
        timer = self.client.timer("some_timer")
        self.assertIsInstance(timer, metrics.Timer)
        self.assertEqual(timer.name, b"namespace.some_timer")

        with self.assertRaises(UnicodeEncodeError):
            self.client.timer("☃")

    def test_make_counter(self):
        counter = self.client.counter("some_counter")
        self.assertIsInstance(counter, metrics.Counter)
        self.assertEqual(counter.name, b"namespace.some_counter")

        with self.assertRaises(UnicodeEncodeError):
            self.client.counter("☃")

    def test_make_gauge(self):
        gauge = self.client.gauge("some_gauge")
        self.assertIsInstance(gauge, metrics.Gauge)
        self.assertEqual(gauge.name, b"namespace.some_gauge")

        with self.assertRaises(UnicodeEncodeError):
            self.client.gauge("☃")

    def test_make_histogram(self):
        histogram = self.client.histogram("some_histogram")
        self.assertIsInstance(histogram, metrics.Histogram)
        self.assertEqual(histogram.name, b"namespace.some_histogram")

        with self.assertRaises(UnicodeEncodeError):
            self.client.gauge("☃")


class ClientTests(unittest.TestCase):
    def test_make_batch(self):
        transport = mock.Mock(spec=metrics.NullTransport)
        client = metrics.Client(transport, "namespace")
        batch = client.batch()

        self.assertIsInstance(batch, metrics.Batch)
        self.assertEqual(batch.namespace, b"namespace")


class BatchTests(unittest.TestCase):
    def setUp(self):
        self.patcher = mock.patch("baseplate.lib.metrics.BufferedTransport", autospec=True)
        self.mock_buffer = self.patcher.start().return_value
        self.mock_transport = mock.Mock(spec=metrics.NullTransport)

        # encode is called here since metrics.Batch is designed to be instantiated
        # by an instance of metrics.Client which encodes the namespace arg
        self.batch = metrics.Batch(self.mock_transport, b"namespace")

    def test_context(self):
        with self.batch as b:
            self.assertEqual(b, self.batch)
        self.assertTrue(self.mock_buffer.flush.called)

    def test_make_counter(self):
        batch_counter = self.batch.counter("some_counter")
        self.assertIsInstance(batch_counter, metrics.BatchCounter)
        expected_counter_name = b"namespace.some_counter"
        self.assertEqual(batch_counter.name, expected_counter_name)
        self.assertEqual(len(self.batch.counters), 1)
        self.assertTrue(expected_counter_name in self.batch.counters)

    def test_get_counter_twice(self):
        counter_name = "some_counter"
        batch_counter = self.batch.counter(counter_name)
        self.assertIsInstance(batch_counter, metrics.BatchCounter)
        expected_counter_name = b"namespace.some_counter"
        self.assertEqual(batch_counter.name, expected_counter_name)

        refetched_batch_counter = self.batch.counter(counter_name)
        self.assertEqual(len(self.batch.counters), 1)
        self.assertTrue(expected_counter_name in self.batch.counters)
        self.assertEqual(refetched_batch_counter, batch_counter)

    @mock.patch("baseplate.lib.metrics.BatchCounter", autospec=True)
    def test_counter_flush(self, MockBatchCounter):
        with self.batch as b:
            batch_counter = b.counter("some_counter")
            batch_counter.increment()
        self.assertTrue(batch_counter.flush.called)

    def test_counters_cleared_after_flush(self):
        self.batch.counter("some_counter").increment()
        self.batch.flush()
        self.mock_buffer.send.assert_called()
        self.mock_buffer.reset_mock()
        self.batch.flush()
        self.mock_buffer.send.assert_not_called()

    def tearDown(self):
        self.patcher.stop()


class TimerTests(unittest.TestCase):
    def setUp(self):
        self.transport = mock.Mock(spec=metrics.NullTransport)

    def test_init_with_tags(self):
        tags = {"test": "true"}
        timer = metrics.Timer(self.transport, b"example", tags)
        self.assertEqual(timer.tags, tags)

    @mock.patch("time.time", autospec=True)
    def test_basic_operation(self, mock_time):
        timer = metrics.Timer(self.transport, b"example")
        self.assertEqual(timer.tags, {})

        with self.assertRaises(Exception):
            timer.stop()

        mock_time.return_value = 1000
        timer.start()
        with self.assertRaises(Exception):
            timer.start()
        self.assertEqual(self.transport.send.call_count, 0)

        mock_time.return_value = 1004
        timer.stop()
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:4000|ms"))

        with self.assertRaises(Exception):
            timer.start()
        with self.assertRaises(Exception):
            timer.stop()

    @mock.patch("time.time", autospec=True)
    def test_context_manager(self, mock_time):
        timer = metrics.Timer(self.transport, b"example")

        mock_time.return_value = 1000
        with timer:
            mock_time.return_value = 1003

        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:3000|ms"))

    def test_send(self):
        timer = metrics.Timer(self.transport, b"example")
        timer.send(3.14)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:3140|ms"))

    def test_send_tagged(self):
        tags = {"test": "true"}
        timer = metrics.Timer(self.transport, b"example", tags)
        timer.send(3.14)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example,test=true:3140|ms"))

    def test_update_tags(self):
        tags = {"test": "true"}
        timer = metrics.Timer(self.transport, b"example", tags)
        new_tags = {"test2": "false"}
        timer.update_tags(new_tags)
        self.assertNotEqual(timer.tags, new_tags)


class CounterTests(unittest.TestCase):
    def setUp(self):
        self.transport = mock.Mock(spec=metrics.NullTransport)

    def test_incr(self):
        counter = metrics.Counter(self.transport, b"example")
        self.assertIsNone(counter.tags)

        counter.increment()
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:1|c"))

        counter.increment(delta=10)
        self.assertEqual(self.transport.send.call_count, 2)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:10|c"))

        counter.increment(delta=-20)
        self.assertEqual(self.transport.send.call_count, 3)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:-20|c"))

        counter.increment(delta=2, sample_rate=0.5)
        self.assertEqual(self.transport.send.call_count, 4)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:2|c|@0.5"))

    def test_decr(self):
        counter = metrics.Counter(self.transport, b"example")

        counter.decrement()
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:-1|c"))

        counter.decrement(delta=3)
        self.assertEqual(self.transport.send.call_count, 2)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:-3|c"))

    def test_send_tagged(self):
        tags = {"test": "true"}
        counter = metrics.Counter(self.transport, b"example", tags)
        counter.send(delta=2, sample_rate=0.5)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example,test=true:2|c|@0.5"))


class BatchCounterTests(unittest.TestCase):
    def setUp(self):
        self.transport = mock.Mock(spec=metrics.NullTransport)

    def test_increment(self):
        batch_counter = metrics.BatchCounter(self.transport, b"example")
        batch_counter.increment()
        self.assertEqual(len(batch_counter.packets), 1)
        self.assertEqual(batch_counter.packets[1.0], 1)
        self.assertFalse(self.transport.send.called)

    def test_multiple_increments(self):
        batch_counter = metrics.BatchCounter(self.transport, b"example")
        batch_counter.increment()
        self.assertFalse(self.transport.send.called)
        batch_counter.increment()
        self.assertEqual(batch_counter.packets[1.0], 2)
        self.assertFalse(self.transport.send.called)

    def test_flush(self):
        batch_counter = metrics.BatchCounter(self.transport, b"example")
        batch_counter.increment()
        batch_counter.increment(3)
        self.assertEqual(batch_counter.packets[1.0], 4)
        batch_counter.flush()
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:4|c"))

    def test_multiple_sample_rates(self):
        batch_counter = metrics.BatchCounter(self.transport, b"example")
        batch_counter.increment()
        batch_counter.increment()
        batch_counter.increment(sample_rate=0.5)
        batch_counter.flush()

        self.assertEqual(len(batch_counter.packets), 2)
        self.assertEqual(batch_counter.packets[1.0], 2)
        self.assertEqual(batch_counter.packets[0.5], 1)
        self.assertEqual(self.transport.send.call_count, 2)
        expected_call_args = [mock.call(b"example:2|c"), mock.call(b"example:1|c|@0.5")]
        for expected in expected_call_args:
            self.assertTrue(expected in self.transport.send.call_args_list)


class GaugeTests(unittest.TestCase):
    def setUp(self):
        self.transport = mock.Mock(spec=metrics.NullTransport)

    def test_replace(self):
        gauge = metrics.Gauge(self.transport, b"example")
        gauge.replace(33)
        self.assertIsNone(gauge.tags)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example:33|g"))

    def test_replace_disallow_negative(self):
        gauge = metrics.Gauge(self.transport, b"example")
        with self.assertRaises(Exception):
            gauge.replace(-2)

    def test_replace_tagged(self):
        tags = {"test": "true"}
        gauge = metrics.Gauge(self.transport, b"example", tags)
        gauge.replace(33)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example,test=true:33|g"))


class HistogramTests(unittest.TestCase):
    def setUp(self):
        self.transport = mock.Mock(spec=metrics.NullTransport)

    def test_log(self):
        histogram = metrics.Histogram(self.transport, b"example_hist")
        histogram.add_sample(33)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example_hist:33|h"))

    def test_log_tagged(self):
        tags = {"test": "true"}
        histogram = metrics.Histogram(self.transport, b"example_hist", tags)
        self.assertEqual(histogram.tags, tags)
        histogram.add_sample(33)
        self.assertEqual(self.transport.send.call_count, 1)
        self.assertEqual(self.transport.send.call_args, mock.call(b"example_hist,test=true:33|h"))


class MakeClientTests(unittest.TestCase):
    def test_no_endpoint(self):
        client = metrics.make_client("namespace", None, log_if_unconfigured=False)
        self.assertIsInstance(client.transport, metrics.NullTransport)

    def test_valid_endpoint(self):
        client = metrics.make_client("namespace", EXAMPLE_ENDPOINT, log_if_unconfigured=False)
        self.assertIsInstance(client.transport, metrics.RawTransport)



import unittest

from unittest import mock

try:
    from sqlalchemy.engine.url import URL
except ImportError:
    raise unittest.SkipTest("sqlalchemy is not installed")

from baseplate.clients.sqlalchemy import engine_from_config
from baseplate.clients.sqlalchemy import SQLAlchemyEngineContextFactory
from baseplate.testing.lib.secrets import FakeSecretsStore

from prometheus_client import REGISTRY
from sqlalchemy.pool import QueuePool


class EngineFromConfigTests(unittest.TestCase):
    def setUp(self):
        secrets = FakeSecretsStore(
            {
                "secrets": {
                    "secret/sql/account": {
                        "type": "credential",
                        "username": "reddit",
                        "password": "password",
                    }
                }
            }
        )
        self.secrets = secrets

    def test_url(self):
        engine = engine_from_config({"database.url": "sqlite://"})
        self.assertEqual(engine.url, URL("sqlite"))

    @mock.patch("baseplate.clients.sqlalchemy.create_engine")
    def test_credentials(self, create_engine_mock):
        engine_from_config(
            {
                "database.url": "postgresql://fizz:buzz@localhost:9000/db",
                "database.credentials_secret": "secret/sql/account",
                "database.pool_recycle": "60",
                "database.pool_size": "10",
                "database.max_overflow": "5",
            },
            self.secrets,
        )
        create_engine_mock.assert_called_once_with(
            URL(
                drivername="postgresql",
                username="reddit",
                password="password",
                host="localhost",
                port="9000",
                database="db",
            ),
            pool_recycle=60,
            pool_size=10,
            max_overflow=5,
        )

    @mock.patch("baseplate.clients.sqlalchemy.create_engine")
    def test_credentials_no_secrets(self, create_engine_mock):
        with self.assertRaises(TypeError):
            engine_from_config(
                {
                    "database.url": "postgresql://fizz:buzz@localhost:9000/db",
                    "database.credentials_secret": "secret/sql/account",
                }
            )
        self.assertEqual(create_engine_mock.call_count, 0)


class EngineContextFactoryTest(unittest.TestCase):
    @mock.patch("baseplate.clients.sqlalchemy.event")  # I'm just ignoring all the events
    def setUp(self, event):
        SQLAlchemyEngineContextFactory.max_connections_gauge.clear()
        SQLAlchemyEngineContextFactory.checked_out_connections_gauge.clear()
        SQLAlchemyEngineContextFactory.latency_seconds.clear()
        SQLAlchemyEngineContextFactory.requests_total.clear()
        SQLAlchemyEngineContextFactory.active_requests.clear()

        engine = mock.MagicMock()
        # engine.execution_options.return_value = mock.MagicMock()
        self.factory = SQLAlchemyEngineContextFactory(engine, "factory_name")

    def test_report_runtime_metrics_prom_no_queue_pool(self):
        batch = mock.MagicMock()

        pool = mock.MagicMock()  # this will fail the isinstance check
        pool.size.return_value = 4
        pool.checkedout.return_value = 12
        self.factory.engine.pool = pool

        self.factory.report_runtime_metrics(batch)

        prom_labels = {"sql_pool": "factory_name"}
        # this serves to prove that we never set these metrics / go down the code path after the isinstance check
        self.assertEqual(REGISTRY.get_sample_value("sql_client_pool_max_size", prom_labels), None)
        self.assertEqual(
            REGISTRY.get_sample_value("sql_client_pool_client_connections", prom_labels),
            None,
        )

    def test_report_runtime_metrics_prom_with_queue_pool(self):
        """
        This test is to prove that we set the metrics correctly when we have a QueuePool

        We are testing that the metrics report what the pool does. Technically the math here
        would imply that there are 0 idle connections since overflow connections are in use.
        """
        batch = mock.MagicMock()

        pool = mock.MagicMock(spec=QueuePool)  # this will pass the isinstance check
        pool.size.return_value = 4
        # includes overflow connections
        pool.checkedout.return_value = 12
        # unused in the test but required for the rest of the report_runtime_metrics call
        pool.overflow.return_value = 16
        # fake the idle connections for testing the metric value
        pool.checkedin.return_value = 2
        self.factory.engine.pool = pool

        self.factory.report_runtime_metrics(batch)

        prom_labels = {"sql_pool": "factory_name"}
        self.assertEqual(REGISTRY.get_sample_value("sql_client_pool_max_size", prom_labels), 4)
        self.assertEqual(
            REGISTRY.get_sample_value("sql_client_pool_active_connections", prom_labels), 12
        )
        self.assertEqual(
            REGISTRY.get_sample_value("sql_client_pool_idle_connections", prom_labels), 2
        )

    def test_on_before_execute(self):
        conn = mock.MagicMock()
        conn.engine.url.host = "test_hostname"
        conn.engine.url.database = "test_database"
        server_span = mock.MagicMock()
        server_span.make_child.side_effect = [
            mock.MagicMock(trace_id="test_span_trace_id", id="test_span_id")
        ]
        conn._execution_options = {
            "context_name": "test_context_name",
            "server_span": server_span,
        }
        self.factory.on_before_execute(
            conn=conn,
            cursor=None,
            statement="",
            parameters=None,
            context=None,
            executemany=False,
        )

        prom_labels = {
            "sql_pool": "factory_name",
            "sql_address": "test_hostname",
            "sql_database": "test_database",
        }

        self.assertEqual(REGISTRY.get_sample_value("sql_client_active_requests", prom_labels), 1)

    def test_on_after_execute(self):
        conn = mock.MagicMock()
        conn.engine.url.host = "test_hostname"
        conn.engine.url.database = "test_database"
        self.factory.on_after_execute(
            conn=conn,
            cursor=None,
            statement="",
            parameters=None,
            context=None,
            executemany=False,
        )

        prom_labels = {
            "sql_pool": "factory_name",
            "sql_address": "test_hostname",
            "sql_database": "test_database",
        }

        self.assertEqual(
            REGISTRY.get_sample_value(
                "sql_client_requests_total", {**prom_labels, "sql_success": "true"}
            ),
            1,
        )
        self.assertEqual(
            REGISTRY.get_sample_value(
                "sql_client_latency_seconds_bucket",
                {**prom_labels, "le": "+Inf", "sql_success": "true"},
            ),
            1,
        )
        self.assertEqual(REGISTRY.get_sample_value("sql_client_active_requests", prom_labels), -1)

    def test_on_error(self):
        exception_context = mock.MagicMock()
        exception_context.connection.engine.url.host = "test_hostname"
        exception_context.connection.engine.url.database = "test_database"
        self.factory.on_error(exception_context)

        prom_labels = {
            "sql_pool": "factory_name",
            "sql_address": "test_hostname",
            "sql_database": "test_database",
        }

        self.assertEqual(
            REGISTRY.get_sample_value(
                "sql_client_requests_total", {**prom_labels, "sql_success": "false"}
            ),
            1,
        )
        self.assertEqual(
            REGISTRY.get_sample_value(
                "sql_client_latency_seconds_bucket",
                {**prom_labels, "le": "+Inf", "sql_success": "false"},
            ),
            1,
        )
        self.assertEqual(REGISTRY.get_sample_value("sql_client_active_requests", prom_labels), -1)



import unittest

from typing import Any

try:
    import rediscluster
except ImportError:
    raise unittest.SkipTest("redis-py-cluster is not installed")

from prometheus_client import REGISTRY

from baseplate.lib.config import ConfigurationError
from baseplate.clients.redis_cluster import cluster_pool_from_config

from baseplate.clients.redis_cluster import ClusterRedisClient
from baseplate import Baseplate
from . import TestBaseplateObserver, get_endpoint_or_skip_container
from baseplate.clients.redis_cluster import ACTIVE_REQUESTS
from baseplate.clients.redis_cluster import LATENCY_SECONDS
from baseplate.clients.redis_cluster import REQUESTS_TOTAL

redis_endpoint = get_endpoint_or_skip_container("redis-cluster-node", 7000)


# This belongs on the unit tests section but the client class attempts to initialise
# the list of nodes when being instantiated so it's simpler to test here with a redis
# cluster available
class ClusterPoolFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            cluster_pool_from_config({})

    def test_basic_url(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        self.assertEqual(pool.nodes.startup_nodes[0]["host"], "redis-cluster-node")
        self.assertEqual(pool.nodes.startup_nodes[0]["port"], "7000")

    def test_timeouts(self):
        pool = cluster_pool_from_config(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.timeout": "30 seconds",
            }
        )

        self.assertEqual(pool.timeout, 30)

    def test_max_connections(self):
        pool = cluster_pool_from_config(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.max_connections": "300",
            }
        )

        self.assertEqual(pool.max_connections, 300)

    def test_max_connections_default(self):
        # https://github.com/Grokzen/redis-py-cluster/issues/435
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        self.assertEqual(pool.max_connections, 50)

    def test_kwargs_passthrough(self):
        pool = cluster_pool_from_config(
            {"rediscluster.url": f"redis://{redis_endpoint}/0"}, example="present"
        )

        self.assertEqual(pool.connection_kwargs["example"], "present")

    def test_alternate_prefix(self):
        pool = cluster_pool_from_config(
            {"noodle.url": f"redis://{redis_endpoint}/0"}, prefix="noodle."
        )
        self.assertEqual(pool.nodes.startup_nodes[0]["host"], "redis-cluster-node")
        self.assertEqual(pool.nodes.startup_nodes[0]["port"], "7000")

    def test_only_primary_available(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})
        node_list = [pool.get_node_by_slot(slot=1, read_command=False) for _ in range(0, 100)]

        # The primary is on port 7000 so that's the only port we expect to see
        self.assertTrue(all(node["port"] == 7000 for node in node_list))

    def test_read_from_replicas(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        node_list = [pool.get_node_by_slot(slot=1, read_command=True) for _ in range(0, 100)]

        # Both replicas and primary are available, so we expect to see some non-primaries here
        self.assertTrue(any(node["port"] != 7000 for node in node_list))


class RedisClusterIntegrationTests(unittest.TestCase):
    def setUp(self):
        self.setup_baseplate_redis(
            {
                "redis_client_name": "test_client",
            },
        )

    def setup_baseplate_redis(self, redis_cluster_kwargs: dict[str, Any] = {}):
        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.timeout": "1 second",
                "rediscluster.max_connections": "4",
            }
        )
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context({"rediscluster": ClusterRedisClient(**redis_cluster_kwargs)})

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_command(self):
        with self.server_span:
            result = self.context.rediscluster.ping()

        self.assertTrue(result)

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "rediscluster.PING")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_error(self):
        with self.server_span:
            with self.assertRaises(rediscluster.RedisClusterException):
                self.context.rediscluster.execute_command("crazycommand")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)

    def test_lock(self):
        with self.server_span:
            with self.context.rediscluster.lock("foo-lock"):
                pass

        server_span_observer = self.baseplate_observer.get_only_child()

        self.assertGreater(len(server_span_observer.children), 0)
        for span_observer in server_span_observer.children:
            self.assertTrue(span_observer.on_start_called)
            self.assertTrue(span_observer.on_finish_called)

    def test_pipeline(self):
        with self.server_span:
            with self.context.rediscluster.pipeline("foo") as pipeline:
                pipeline.set("foo", "bar")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.delete("foo")
                pipeline.execute()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "rediscluster.pipeline_foo")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_metrics(self):
        client_name = "redis_test"
        for client_name_kwarg_name in [
            "redis_client_name",
            "client_name",
        ]:
            with self.subTest():
                self.setup_baseplate_redis(
                    redis_cluster_kwargs={
                        client_name_kwarg_name: client_name,
                    },
                )
                # Clear preometheus metrics
                ACTIVE_REQUESTS.clear()
                REQUESTS_TOTAL.clear()
                LATENCY_SECONDS.clear()
                expected_labels = {
                    "redis_client_name": client_name,
                    "redis_type": "cluster",
                    "redis_command": "SET",
                    "redis_database": "0",
                }
                with self.server_span:
                    self.context.rediscluster.set("prometheus", "rocks")

                request_labels = {**expected_labels, "redis_success": "true"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", request_labels)
                    == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'set' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**request_labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'set' latency request"
                assert (
                    REGISTRY.get_sample_value(ACTIVE_REQUESTS._name, {**expected_labels}) == 0.0
                ), "Should have 0 (and not None) active requests"

    def test_pipeline_metrics(self):
        client_name = "test_client"
        for client_name_kwarg_name in [
            "redis_client_name",
            "client_name",
        ]:
            with self.subTest():
                self.setup_baseplate_redis(
                    redis_cluster_kwargs={
                        client_name_kwarg_name: client_name,
                    },
                )
                # Clear preometheus metrics
                ACTIVE_REQUESTS.clear()
                REQUESTS_TOTAL.clear()
                LATENCY_SECONDS.clear()
                expected_labels = {
                    "redis_client_name": client_name,
                    "redis_type": "cluster",
                    "redis_command": "pipeline",
                    "redis_database": "0",
                }
                with self.server_span:
                    with self.context.rediscluster.pipeline("foo") as pipeline:
                        pipeline.set("foo", "bar")
                        pipeline.get("foo")
                        pipeline.get("foo")
                        pipeline.get("foo")
                        pipeline.get("foo")
                        pipeline.get("foo")
                        pipeline.delete("foo")
                        pipeline.execute()

                request_labels = {**expected_labels, "redis_success": "true"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", request_labels)
                    == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'set' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**request_labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'set' latency request"
                assert (
                    REGISTRY.get_sample_value(ACTIVE_REQUESTS._name, {**expected_labels}) == 0.0
                ), "Should have 0 (and not None) active requests"



import unittest

from baseplate.lib.secrets import CorruptSecretError
from baseplate.lib.secrets import CredentialSecret
from baseplate.lib.secrets import SecretNotFoundError
from baseplate.lib.secrets import secrets_store_from_config
from baseplate.lib.secrets import SecretsNotAvailableError
from baseplate.lib.secrets import SecretsStore
from baseplate.testing.lib.file_watcher import FakeFileWatcher


class StoreTests(unittest.TestCase):
    def setUp(self):
        self.fake_filewatcher = FakeFileWatcher()
        self.store = SecretsStore("/whatever")
        self.store._filewatcher = self.fake_filewatcher

    def test_file_not_found(self):
        with self.assertRaises(SecretsNotAvailableError):
            self.store.get_raw("test")

    def test_vault_info(self):
        self.fake_filewatcher.data = {
            "secrets": {},
            "vault": {"token": "test", "url": "http://vault.example.com:8200/"},
        }

        self.assertEqual(self.store.get_vault_token(), "test")
        self.assertEqual(self.store.get_vault_url(), "http://vault.example.com:8200/")

    def test_raw_secrets(self):
        self.fake_filewatcher.data = {
            "secrets": {"test": {"something": "exists"}},
            "vault": {"token": "test", "url": "http://vault.example.com:8200/"},
        }

        self.assertEqual(self.store.get_raw("test"), {"something": "exists"})

        with self.assertRaises(SecretNotFoundError):
            self.store.get_raw("test_missing")

    def test_simple_secrets(self):
        self.fake_filewatcher.data = {
            "secrets": {
                "test": {"type": "simple", "value": "easy"},
                "test_base64": {"type": "simple", "value": "aHVudGVyMg==", "encoding": "base64"},
                "test_unknown_encoding": {
                    "type": "simple",
                    "value": "sdlfkj",
                    "encoding": "mystery",
                },
                "test_not_simple": {"something": "else"},
                "test_no_value": {"type": "simple"},
                "test_bad_base64": {"type": "simple", "value": "aHVudGVyMg", "encoding": "base64"},
            },
            "vault": {"token": "test", "url": "http://vault.example.com:8200/"},
        }

        self.assertEqual(self.store.get_simple("test"), b"easy")
        self.assertEqual(self.store.get_simple("test_base64"), b"hunter2")

        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("test_unknown_encoding")

        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("test_not_simple")

        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("test_no_value")

        with self.assertRaises(CorruptSecretError):
            self.store.get_simple("test_bad_base64")

    def test_versioned_secrets(self):
        self.fake_filewatcher.data = {
            "secrets": {
                "test": {"type": "versioned", "current": "easy"},
                "test_base64": {
                    "type": "versioned",
                    "previous": "aHVudGVyMQ==",
                    "current": "aHVudGVyMg==",
                    "next": "aHVudGVyMw==",
                    "encoding": "base64",
                },
                "test_unknown_encoding": {
                    "type": "versioned",
                    "current": "sdlfkj",
                    "encoding": "mystery",
                },
                "test_not_versioned": {"something": "else"},
                "test_no_value": {"type": "versioned"},
                "test_bad_base64": {"type": "simple", "value": "aHVudGVyMg", "encoding": "base64"},
            },
            "vault": {"token": "test", "url": "http://vault.example.com:8200/"},
        }

        simple = self.store.get_versioned("test")
        self.assertEqual(simple.current, b"easy")
        self.assertEqual(list(simple.all_versions), [b"easy"])

        encoded = self.store.get_versioned("test_base64")
        self.assertEqual(encoded.previous, b"hunter1")
        self.assertEqual(encoded.current, b"hunter2")
        self.assertEqual(encoded.next, b"hunter3")
        self.assertEqual(list(encoded.all_versions), [b"hunter2", b"hunter1", b"hunter3"])

        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("test_unknown_encoding")

        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("test_not_versioned")

        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("test_no_value")

        with self.assertRaises(CorruptSecretError):
            self.store.get_versioned("test_bad_base64")

    def test_credential_secrets(self):
        self.fake_filewatcher.data = {
            "secrets": {
                "test": {"type": "credential", "username": "user", "password": "password"},
                "test_identity": {
                    "type": "credential",
                    "username": "spez",
                    "password": "hunter2",
                    "encoding": "identity",
                },
                "test_base64": {
                    "type": "credential",
                    "username": "foo",
                    "password": "aHVudGVyMg==",
                    "encoding": "base64",
                },
                "test_unknown_encoding": {
                    "type": "credential",
                    "username": "fizz",
                    "password": "buzz",
                    "encoding": "something",
                },
                "test_not_credentials": {"type": "versioned", "current": "easy"},
                "test_no_values": {"type": "credential"},
                "test_no_username": {"type": "credential", "password": "password"},
                "test_no_password": {"type": "credential", "username": "user"},
                "test_bad_type": {"type": "credential", "username": "user", "password": 100},
            },
            "vault": {"token": "test", "url": "http://vault.example.com:8200/"},
        }

        self.assertEqual(self.store.get_credentials("test"), CredentialSecret("user", "password"))
        self.assertEqual(
            self.store.get_credentials("test_identity"), CredentialSecret("spez", "hunter2")
        )

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_base64")

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_unknown_encoding")

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_not_credentials")

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_no_values")

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_no_username")

        with self.assertRaises(CorruptSecretError):
            self.store.get_credentials("test_no_password")


class StoreFromConfigTests(unittest.TestCase):
    def test_make_store(self):
        secrets = secrets_store_from_config({"secrets.path": "/tmp/test"})
        self.assertIsInstance(secrets, SecretsStore)

    def test_prefix(self):
        secrets = secrets_store_from_config(
            {"secrets.path": "/tmp/test", "test_secrets.path": "/tmp/secrets"},
            prefix="test_secrets.",
        )
        self.assertIsInstance(secrets, SecretsStore)
        self.assertEqual(secrets._filewatcher._path, "/tmp/secrets")



from typing import Any
from typing import Dict

import gevent
import pytest
import sentry_sdk

from baseplate import Baseplate
from baseplate.observers.sentry import _SentryUnhandledErrorReporter
from baseplate.observers.sentry import init_sentry_client_from_config
from baseplate.observers.sentry import SentryBaseplateObserver


class FakeTransport:
    def __init__(self):
        self.events = []

    def __call__(self, event: Dict[str, Any]) -> None:
        self.events.append(event)


@pytest.fixture
def sentry_transport():
    return FakeTransport()


@pytest.fixture(autouse=True)
def init_sentry_client(sentry_transport):
    try:
        init_sentry_client_from_config({"sentry.dsn": "foo"}, transport=sentry_transport)
        yield
    finally:
        sentry_sdk.init()  # shut everything down


@pytest.fixture
def baseplate_app():
    baseplate = Baseplate()
    baseplate.register(SentryBaseplateObserver())
    return baseplate


def test_no_event_when_nothing_wrong(baseplate_app, sentry_transport):
    with baseplate_app.server_context("test"):
        pass

    assert not sentry_transport.events


def test_event_when_exception(baseplate_app, sentry_transport):
    with pytest.raises(ValueError):
        with baseplate_app.server_context("test"):
            raise ValueError("oops")

    assert len(sentry_transport.events) == 1
    event = sentry_transport.events[0]
    assert event["exception"]["values"][0]["type"] == "ValueError"


def test_tags(baseplate_app, sentry_transport):
    with pytest.raises(ValueError):
        with baseplate_app.server_context("test") as context:
            context.span.set_tag("foo", "bar")
            raise ValueError("oops")

    assert len(sentry_transport.events) == 1
    event = sentry_transport.events[0]
    assert event["exception"]["values"][0]["type"] == "ValueError"
    assert event["tags"]["foo"] == "bar"

    with pytest.raises(ValueError):
        with baseplate_app.server_context("test") as context:
            context.span.set_tag("different-tag", "foo")
            raise ValueError("oops")

    assert len(sentry_transport.events) == 2
    event = sentry_transport.events[1]
    assert event["exception"]["values"][0]["type"] == "ValueError"
    assert event["tags"]["different-tag"] == "foo"
    assert "foo" not in event["tags"]


def test_logs(baseplate_app, sentry_transport):
    with pytest.raises(ValueError):
        with baseplate_app.server_context("test") as context:
            context.span.log("foo-category", "bar-log-entry")
            raise ValueError("oops")

    assert len(sentry_transport.events) == 1
    event = sentry_transport.events[0]
    assert event["exception"]["values"][0]["type"] == "ValueError"

    last_breadcrumb = event["breadcrumbs"]["values"][-1]
    assert last_breadcrumb["category"] == "foo-category"
    assert last_breadcrumb["message"] == "bar-log-entry"


def test_ignored_exception_ignored(baseplate_app, sentry_transport):
    with pytest.raises(ConnectionError):
        with baseplate_app.server_context("test"):
            raise ConnectionError("oops")

    assert not sentry_transport.events


def test_unhandled_error_reporter(sentry_transport):
    def raise_unhandled_error():
        raise KeyError("foo")

    try:
        _SentryUnhandledErrorReporter.install()

        greenlet = gevent.spawn(raise_unhandled_error)
        greenlet.join()
    finally:
        _SentryUnhandledErrorReporter.uninstall()

    assert len(sentry_transport.events) == 1
    event = sentry_transport.events[0]
    assert event["exception"]["values"][0]["type"] == "KeyError"


def test_unhandled_error_reporter_server_timeout(sentry_transport):
    def raise_unhandled_error():
        from baseplate.observers.timeout import ServerTimeout

        raise ServerTimeout("blah", 1.0, debug=False)

    try:
        _SentryUnhandledErrorReporter.install()

        greenlet = gevent.spawn(raise_unhandled_error)
        greenlet.join()
    finally:
        _SentryUnhandledErrorReporter.uninstall()

    assert not sentry_transport.events



#
# Autogenerated by Thrift Compiler (0.14.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:slots
#
import logging
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TFrozenDict
from thrift.Thrift import TMessageType
from thrift.Thrift import TProcessor
from thrift.Thrift import TType
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

from .ttypes import *

all_structs = []


class Iface(object):
    def example(self):
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def example(self):
        self.send_example()
        return self.recv_example()

    def send_example(self):
        self._oprot.writeMessageBegin("example", TMessageType.CALL, self._seqid)
        args = example_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_example(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = example_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.exc is not None:
            raise result.exc
        if result.err is not None:
            raise result.err
        raise TApplicationException(
            TApplicationException.MISSING_RESULT, "example failed: unknown result"
        )


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["example"] = Processor.process_example
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(
                TApplicationException.UNKNOWN_METHOD, "Unknown function %s" % (name)
            )
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_example(self, seqid, iprot, oprot):
        args = example_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = example_result()
        try:
            result.success = self._handler.example()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except ExpectedException as exc:
            msg_type = TMessageType.REPLY
            result.exc = exc
        except baseplate.thrift.ttypes.Error as err:
            msg_type = TMessageType.REPLY
            result.err = err
        except TApplicationException as ex:
            logging.exception("TApplication exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception("Unexpected exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, "Internal error")
        oprot.writeMessageBegin("example", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES


class example_args(object):

    __slots__ = ()

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("example_args")
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(example_args)
example_args.thrift_spec = ()


class example_result(object):
    """
    Attributes:
     - success
     - exc
     - err

    """

    __slots__ = (
        "success",
        "exc",
        "err",
    )

    def __init__(
        self,
        success=None,
        exc=None,
        err=None,
    ):
        self.success = success
        self.exc = exc
        self.err = err

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.exc = ExpectedException.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.err = baseplate.thrift.ttypes.Error.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("example_result")
        if self.success is not None:
            oprot.writeFieldBegin("success", TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        if self.exc is not None:
            oprot.writeFieldBegin("exc", TType.STRUCT, 1)
            self.exc.write(oprot)
            oprot.writeFieldEnd()
        if self.err is not None:
            oprot.writeFieldBegin("err", TType.STRUCT, 2)
            self.err.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(example_result)
example_result.thrift_spec = (
    (
        0,
        TType.BOOL,
        "success",
        None,
        None,
    ),  # 0
    (
        1,
        TType.STRUCT,
        "exc",
        [ExpectedException, None],
        None,
    ),  # 1
    (
        2,
        TType.STRUCT,
        "err",
        [baseplate.thrift.ttypes.Error, None],
        None,
    ),  # 2
)
fix_spec(all_structs)
del all_structs



import builtins
import unittest

from unittest import mock

try:
    import pymemcache
except ImportError:
    raise unittest.SkipTest("pymemcache is not installed")
else:
    del pymemcache

from prometheus_client import REGISTRY
from baseplate.lib.config import ConfigurationError
from baseplate.clients.memcache import pool_from_config
from baseplate.clients.memcache import MonitoredMemcacheConnection
from baseplate.clients.memcache import lib as memcache_lib


class PrometheusInstrumentationTests(unittest.TestCase):
    def setUp(self):
        # stub out the server span, each call to MonitoredMemcacheConnection
        # will create a new span - but we don't use that for prom
        span_inner = mock.Mock()
        span_inner.__enter__ = mock.Mock(return_value=mock.Mock())
        span_inner.__exit__ = mock.Mock(return_value=None)
        self.span = mock.Mock()
        self.span.make_child = lambda trace_name: span_inner

    # test the @_prom_instrument decorator by calling one of the decorated functions
    def test_prometheus_decorator(self):
        # stub out the client, we only use the server name
        pooled_client = mock.Mock()
        pooled_client.server = "server name"

        prom_labels = {
            "memcached_address": "server name",
            "memcached_command": "incr",
        }

        mmc = MonitoredMemcacheConnection("test", self.span, pooled_client)
        mmc.incr("blah", 1)

        assert (
            REGISTRY.get_sample_value(
                "memcached_client_active_requests",
                prom_labels,
            )
            == 0
        )
        assert (
            REGISTRY.get_sample_value(
                "memcached_client_requests_total",
                {
                    **prom_labels,
                    "memcached_success": "true",
                },
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "memcached_client_latency_seconds_bucket",
                {
                    **prom_labels,
                    "memcached_success": "true",
                    "le": "+Inf",
                },
            )
            == 1
        )

    def test_prometheus_decorator_failing(self):
        class FailClient(mock.Mock):
            server = "hello"

            def incr(self, _a, _b, noreply=None):
                raise TypeError

        fail_mmc = MonitoredMemcacheConnection("test", self.span, FailClient())
        self.assertRaises(TypeError, fail_mmc.incr, "aaa", 1)

        assert (
            REGISTRY.get_sample_value(
                "memcached_client_requests_total",
                {
                    "memcached_address": "hello",
                    "memcached_command": "incr",
                    "memcached_success": "false",
                },
            )
            == 1
        )


class PoolFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            pool_from_config({})

    def test_basic_url(self):
        pool = pool_from_config({"memcache.endpoint": "localhost:1234"})

        self.assertEqual(pool.server[0], "localhost")
        self.assertEqual(pool.server[1], 1234)

    def test_timeouts(self):
        pool = pool_from_config(
            {
                "memcache.endpoint": "localhost:1234",
                "memcache.timeout": "1.23",
                "memcache.connect_timeout": "4.56",
            }
        )

        self.assertEqual(pool.timeout, 1.23)
        self.assertEqual(pool.connect_timeout, 4.56)

    def test_max_connections(self):
        pool = pool_from_config(
            {"memcache.endpoint": "localhost:1234", "memcache.max_pool_size": "300"}
        )

        self.assertEqual(pool.client_pool.max_size, 300)

    def test_alternate_prefix(self):
        pool_from_config({"noodle.endpoint": "localhost:1234"}, prefix="noodle.")

    def test_nodelay(self):
        pool = pool_from_config(
            {"memcache.endpoint": "localhost:1234", "memcache.no_delay": "False"}
        )
        self.assertEqual(pool.no_delay, False)


class SerdeTests(unittest.TestCase):
    def test_serialize_str(self):
        dump_no_compress = memcache_lib.make_dump_and_compress_fn()
        value, flags = dump_no_compress(key="key", value="val")
        self.assertEqual(value, b"val")
        self.assertEqual(flags, 0)

    def test_serialize_bytes(self):
        dump_no_compress = memcache_lib.make_dump_and_compress_fn()
        value, flags = dump_no_compress(key="key", value=b"val")
        self.assertEqual(value, b"val")
        self.assertEqual(flags, 0)

    def test_serialize_int(self):
        dump_no_compress = memcache_lib.make_dump_and_compress_fn()
        value, flags = dump_no_compress(key="key", value=100)
        self.assertEqual(value, b"100")
        self.assertEqual(flags, memcache_lib.Flags.INTEGER)

    def test_serialize_other(self):
        json_patch = mock.patch.object(memcache_lib, "json")
        json = json_patch.start()
        self.addCleanup(json_patch.stop)

        json.dumps.return_value = "expected"

        dump_no_compress = memcache_lib.make_dump_and_compress_fn()
        value, flags = dump_no_compress(key="key", value=("stuff", 1, False))

        json.dumps.assert_called_with(("stuff", 1, False))
        self.assertEqual(value, b"expected")
        self.assertEqual(flags, memcache_lib.Flags.JSON)

    def test_deserialize_str(self):
        value = memcache_lib.decompress_and_load(key="key", serialized="val", flags=0)
        self.assertEqual(value, "val")

    def test_deserialize_bytes(self):
        value = memcache_lib.decompress_and_load(key="key", serialized=b"val", flags=0)
        self.assertEqual(value, b"val")

    def test_deserialize_int(self):
        value = memcache_lib.decompress_and_load(
            key="key", serialized="100", flags=memcache_lib.Flags.INTEGER
        )
        self.assertEqual(value, 100)
        self.assertTrue(isinstance(value, int))

    def test_deserialize_long(self):
        value = memcache_lib.decompress_and_load(
            key="key", serialized="100", flags=memcache_lib.Flags.LONG
        )

        self.assertEqual(value, 100)
        self.assertTrue(isinstance(value, int))

    def test_deserialize_other(self):
        json_patch = mock.patch.object(memcache_lib, "json")
        json = json_patch.start()
        self.addCleanup(json_patch.stop)

        expected_value = object()
        json.loads.return_value = expected_value

        value = memcache_lib.decompress_and_load(
            key="key", serialized=b"garbage", flags=memcache_lib.Flags.JSON
        )

        json.loads.assert_called_with(b"garbage")
        self.assertEqual(value, expected_value)

    def test_serialize_no_compress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        dump_no_compress = memcache_lib.make_dump_and_compress_fn(
            min_compress_length=0  # disable compression
        )
        value, flags = dump_no_compress(key="key", value="simple string")
        self.assertEqual(value, b"simple string")
        self.assertEqual(flags, 0)
        zlib.compress.assert_not_called()

    def test_serialize_compress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_value = object()
        zlib.compress.return_value = expected_value

        json_and_compress = memcache_lib.make_dump_and_compress_fn(
            min_compress_length=1, compress_level=1
        )
        value, flags = json_and_compress(key="key", value="simple string")
        self.assertEqual(value, expected_value)
        self.assertEqual(flags, memcache_lib.Flags.ZLIB)
        zlib.compress.assert_called_with(b"simple string", 1)

    def test_deserialize_no_decompress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        value = memcache_lib.decompress_and_load(key="key", serialized="stuff", flags=0)
        self.assertEqual(value, "stuff")
        zlib.decompress.assert_not_called()

    def test_deserialize_decompress_str(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_value = object()
        zlib.decompress.return_value = expected_value

        flags = 0 | memcache_lib.Flags.ZLIB
        value = memcache_lib.decompress_and_load(key="key", serialized=b"nonsense", flags=flags)
        self.assertEqual(value, expected_value)
        zlib.decompress.assert_called_with(b"nonsense")

    def test_deserialize_decompress_int(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        int_patch = mock.patch.object(builtins, "int")
        int_cls = int_patch.start()
        self.addCleanup(int_patch.stop)

        expected_int_value = object()
        int_cls.return_value = expected_int_value
        flags = memcache_lib.Flags.INTEGER | memcache_lib.Flags.ZLIB
        value = memcache_lib.decompress_and_load(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        int_cls.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_int_value)

    def test_deserialize_decompress_long(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        long_patch = mock.patch.object(memcache_lib, "int")
        long_cls = long_patch.start()
        self.addCleanup(long_patch.stop)

        expected_long_value = object()
        long_cls.return_value = expected_long_value
        flags = memcache_lib.Flags.LONG | memcache_lib.Flags.ZLIB
        value = memcache_lib.decompress_and_load(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        long_cls.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_long_value)

    def test_deserialize_decompress_unjson(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        json_patch = mock.patch.object(memcache_lib, "json")
        json = json_patch.start()
        self.addCleanup(json_patch.stop)

        expected_json_value = object()
        json.loads.return_value = expected_json_value

        flags = memcache_lib.Flags.JSON | memcache_lib.Flags.ZLIB
        value = memcache_lib.decompress_and_load(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        json.loads.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_json_value)


class R2CompatSerdeTests(unittest.TestCase):
    def test_serialize_str(self):
        pickle_no_compress = memcache_lib.make_pickle_and_compress_fn()
        value, flags = pickle_no_compress(key="key", value="val")
        self.assertEqual(value, b"val")
        self.assertEqual(flags, 0)

    def test_serialize_bytes(self):
        pickle_no_compress = memcache_lib.make_pickle_and_compress_fn()
        value, flags = pickle_no_compress(key="key", value=b"val")
        self.assertEqual(value, b"val")
        self.assertEqual(flags, 0)

    def test_serialize_int(self):
        pickle_no_compress = memcache_lib.make_pickle_and_compress_fn()
        value, flags = pickle_no_compress(key="key", value=100)
        self.assertEqual(value, b"100")
        self.assertEqual(flags, memcache_lib.PickleFlags.INTEGER)

    def test_serialize_other(self):
        pickle_patch = mock.patch.object(memcache_lib, "pickle")
        pickle = pickle_patch.start()
        self.addCleanup(pickle_patch.stop)

        expected_value = object()
        pickle.dumps.return_value = expected_value

        pickle_no_compress = memcache_lib.make_pickle_and_compress_fn()
        value, flags = pickle_no_compress(key="key", value=("stuff", 1, False))

        pickle.dumps.assert_called_with(("stuff", 1, False), protocol=2)
        self.assertEqual(value, expected_value)
        self.assertEqual(flags, memcache_lib.PickleFlags.PICKLE)

    def test_deserialize_str(self):
        value = memcache_lib.decompress_and_unpickle(key="key", serialized="val", flags=0)
        self.assertEqual(value, "val")

    def test_deserialize_bytes(self):
        value = memcache_lib.decompress_and_unpickle(key="key", serialized=b"val", flags=0)
        self.assertEqual(value, b"val")

    def test_deserialize_int(self):
        value = memcache_lib.decompress_and_unpickle(
            key="key", serialized="100", flags=memcache_lib.PickleFlags.INTEGER
        )
        self.assertEqual(value, 100)
        self.assertTrue(isinstance(value, int))

    def test_deserialize_long(self):
        value = memcache_lib.decompress_and_unpickle(
            key="key", serialized="100", flags=memcache_lib.PickleFlags.LONG
        )

        self.assertEqual(value, 100)
        self.assertTrue(isinstance(value, int))

    def test_deserialize_other(self):
        pickle_patch = mock.patch.object(memcache_lib, "pickle")
        pickle = pickle_patch.start()
        self.addCleanup(pickle_patch.stop)

        expected_value = object()
        pickle.loads.return_value = expected_value

        value = memcache_lib.decompress_and_unpickle(
            key="key", serialized="garbage", flags=memcache_lib.PickleFlags.PICKLE
        )

        pickle.loads.assert_called_with("garbage")
        self.assertEqual(value, expected_value)

    def test_serialize_no_compress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        pickle_no_compress = memcache_lib.make_pickle_and_compress_fn(
            min_compress_length=0  # disable compression
        )
        value, flags = pickle_no_compress(key="key", value="simple string")
        self.assertEqual(value, b"simple string")
        self.assertEqual(flags, 0)
        zlib.compress.assert_not_called()

    def test_serialize_compress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_value = object()
        zlib.compress.return_value = expected_value

        pickle_and_compress = memcache_lib.make_pickle_and_compress_fn(
            min_compress_length=1, compress_level=1
        )
        value, flags = pickle_and_compress(key="key", value="simple string")
        self.assertEqual(value, expected_value)
        self.assertEqual(flags, memcache_lib.PickleFlags.ZLIB)
        zlib.compress.assert_called_with(b"simple string", 1)

    def test_deserialize_no_decompress(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        value = memcache_lib.decompress_and_unpickle(key="key", serialized="stuff", flags=0)
        self.assertEqual(value, "stuff")
        zlib.decompress.assert_not_called()

    def test_deserialize_decompress_str(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_value = object()
        zlib.decompress.return_value = expected_value

        flags = 0 | memcache_lib.PickleFlags.ZLIB
        value = memcache_lib.decompress_and_unpickle(key="key", serialized="nonsense", flags=flags)
        self.assertEqual(value, expected_value)
        zlib.decompress.assert_called_with("nonsense")

    def test_deserialize_decompress_int(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        int_patch = mock.patch.object(builtins, "int")
        int_cls = int_patch.start()
        self.addCleanup(int_patch.stop)

        expected_int_value = object()
        int_cls.return_value = expected_int_value
        flags = memcache_lib.PickleFlags.INTEGER | memcache_lib.PickleFlags.ZLIB
        value = memcache_lib.decompress_and_unpickle(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        int_cls.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_int_value)

    def test_deserialize_decompress_long(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        long_patch = mock.patch.object(memcache_lib, "int")
        long_cls = long_patch.start()
        self.addCleanup(long_patch.stop)

        expected_long_value = object()
        long_cls.return_value = expected_long_value
        flags = memcache_lib.PickleFlags.LONG | memcache_lib.PickleFlags.ZLIB
        value = memcache_lib.decompress_and_unpickle(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        long_cls.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_long_value)

    def test_deserialize_decompress_unpickle(self):
        zlib_patch = mock.patch.object(memcache_lib, "zlib")
        zlib = zlib_patch.start()
        self.addCleanup(zlib_patch.stop)

        expected_zlib_value = object()
        zlib.decompress.return_value = expected_zlib_value

        pickle_patch = mock.patch.object(memcache_lib, "pickle")
        pickle = pickle_patch.start()
        self.addCleanup(pickle_patch.stop)

        expected_pickle_value = object()
        pickle.loads.return_value = expected_pickle_value

        flags = memcache_lib.PickleFlags.PICKLE | memcache_lib.PickleFlags.ZLIB
        value = memcache_lib.decompress_and_unpickle(key="key", serialized="nonsense", flags=flags)
        zlib.decompress.assert_called_with("nonsense")
        pickle.loads.assert_called_with(expected_zlib_value)
        self.assertEqual(value, expected_pickle_value)



"""Internal library helpers."""
import functools
import inspect
import warnings

from typing import Any
from typing import Callable
from typing import Generic
from typing import Type
from typing import TypeVar


def warn_deprecated(message: str) -> None:
    """Emit a deprecation warning from the caller.

    The stacktrace for the warning will point to the place where the function
    calling this was called, rather than in baseplate. This allows the user to
    most easily see where in _their_ code the deprecation is coming from.

    """
    warnings.warn(message, DeprecationWarning, stacklevel=3)


T = TypeVar("T")
R = TypeVar("R")


# cached_property is a renamed copy of pyramid.decorator.reify
# see COPYRIGHT for full license information
class cached_property(Generic[R]):
    """Like @property but the function will only be called once per instance.

    When used as a method decorator, this will act like @property but instead
    of calling the function each time the attribute is accessed, instead it
    will only call it on first access and then replace itself on the instance
    with the return value of the first call.

    """

    def __init__(self, wrapped: Callable[[Any], R]):
        self.wrapped = wrapped
        functools.update_wrapper(self, wrapped)

    def __get__(self, instance: T, owner: Type[Any]) -> R:
        if instance is None:
            return self
        ret = self.wrapped(instance)
        setattr(instance, self.wrapped.__name__, ret)
        return ret


class UnknownCallerError(Exception):
    def __init__(self) -> None:
        super().__init__("Could not determine calling module's name")


def get_calling_module_name() -> str:
    module = inspect.getmodule(inspect.stack()[2].frame)
    if not module:
        raise UnknownCallerError
    return module.__name__



from baseplate import Baseplate
from baseplate.clients.sqlalchemy import SQLAlchemySession
from baseplate.frameworks.pyramid import BaseplateConfigurator
from pyramid.config import Configurator
from pyramid.view import view_config


@view_config(route_name="hello_world", renderer="json")
def hello_world(request):
    result = request.db.execute("SELECT date('now');")
    return {"Hello": "World", "Now": result.scalar()}


def make_wsgi_app(app_config):
    baseplate = Baseplate(app_config)
    baseplate.configure_observers()
    baseplate.configure_context({"db": SQLAlchemySession()})

    configurator = Configurator(settings=app_config)
    configurator.include(BaseplateConfigurator(baseplate).includeme)
    configurator.add_route("hello_world", "/", request_method="GET")
    configurator.scan()
    return configurator.make_wsgi_app()



from pymemcache.client import PooledClient

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.clients.memcache import MemcacheContextFactory
from baseplate.clients.memcache import MonitoredMemcacheConnection
from baseplate.lib.ratelimit.backends import _get_current_bucket
from baseplate.lib.ratelimit.backends import RateLimitBackend


class MemcacheRateLimitBackendContextFactory(ContextFactory):
    """MemcacheRateLimitBackend context factory.

    :param memcache_pool: The memcache pool to back this ratelimit.
    :param prefix: A prefix to add to keys during rate limiting.
        This is useful if you will have two different rate limiters that will
        receive the same keys.

    """

    def __init__(self, memcache_pool: PooledClient, prefix: str = "rl:"):
        self.memcache_context_factory = MemcacheContextFactory(memcache_pool)
        self.prefix = prefix

    def make_object_for_context(self, name: str, span: Span) -> "MemcacheRateLimitBackend":
        memcache = self.memcache_context_factory.make_object_for_context(name, span)
        return MemcacheRateLimitBackend(memcache, prefix=self.prefix)


class MemcacheRateLimitBackend(RateLimitBackend):
    """A Memcache backend for rate limiting.

    :param memcache: A memcached connection.
    :param prefix: A prefix to add to keys during rate limiting.
        This is useful if you will have two different rate limiters that will
        receive the same keys.

    """

    def __init__(self, memcache: MonitoredMemcacheConnection, prefix: str = "rl:"):
        self.memcache = memcache
        self.prefix = prefix

    def consume(self, key: str, amount: int, allowance: int, interval: int) -> bool:
        """Consume the given `amount` from the allowance for the given `key`.

        This will return true if the `key` remains below the `allowance`
        after consuming the given `amount`.

        :param key: The name of the rate limit bucket to consume from.
        :param amount: The amount to consume from the rate limit bucket.
        :param allowance: The maximum allowance for the rate limit bucket.
        :param interval: The interval to reset the allowance.

        """
        current_bucket = _get_current_bucket(interval)
        key = self.prefix + key + current_bucket
        ttl = interval * 2
        self.memcache.add(key, 0, expire=ttl)
        # `incr` will return None if we experience a delay after the prior
        # `add` call that causes the ttl to expire. We default to `amount` in
        # this case.
        count = self.memcache.incr(key, amount) or amount
        return count <= allowance



import base64
import ipaddress
import sys
import time

from typing import Any
from typing import Optional
from typing import Type
from typing import Union

from advocate import AddrValidator
from advocate import ValidatingHTTPAdapter
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from requests import PreparedRequest
from requests import Request
from requests import Response
from requests import Session
from requests.adapters import HTTPAdapter

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.prometheus_metrics import getHTTPSuccessLabel


def http_adapter_from_config(
    app_config: config.RawConfig, prefix: str, **kwargs: Any
) -> HTTPAdapter:
    """Make an HTTPAdapter from a configuration dictionary.

    The keys useful to :py:func:`http_adapter_from_config` should be prefixed,
    e.g. ``http.pool_connections``, ``http.max_retries``, etc. The ``prefix``
    argument specifies the prefix used. Each key is mapped to a corresponding
    keyword argument on the :py:class:`~requests.adapters.HTTPAdapter`
    constructor.

    Supported keys:

    * ``pool_connections``: The number of connections to cache (default: 10).
    * ``pool_maxsize``: The maximum number of connections to keep in the pool
      (default: 10).
    * ``max_retries``: How many times to retry DNS lookups or connection
      attempts, but never sending data (default: 0).
    * ``pool_block``: Whether the connection pool will block when trying to get
      a connection (default: false).

    Additionally, the rules for Advocate's address filtering can be configured
    with the ``filter`` sub-keys:

    * ``filter.ip_allowlist``: A comma-delimited list of IP addresses (1.2.3.4)
        or CIDR-notation (1.2.3.0/24) ranges that the client can always connect to
        (default: anything not on the local network).
    * ``filter.ip_denylist``: A comma-delimited list of IP addresses or
        CIDR-notation ranges the client may never connect to (default: the local network).
    * ``filter.port_allowlist``: A comma-delimited list of TCP port numbers
        that the client can connect to (default: 80, 8080, 443, 8443, 8000).
    * ``filter.port_denylist``: A comma-delimited list of TCP port numbers that
        the client may never connect to (default: none).
    * ``filter.hostname_denylist``: A comma-delimited list of hostnames that
        the client may never connect to (default: none).
    * ``filter.allow_ipv6``: Should the client be allowed to connect to IPv6
        hosts? (default: false, note: IPv6 is tricky to apply filtering rules
        comprehensively to).

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "pool_connections": config.Optional(config.Integer, default=10),
            "pool_maxsize": config.Optional(config.Integer, default=10),
            "max_retries": config.Optional(config.Integer, default=0),
            "pool_block": config.Optional(config.Boolean, default=False),
            "filter": {
                "ip_allowlist": config.Optional(config.TupleOf(ipaddress.ip_network)),
                "ip_denylist": config.Optional(config.TupleOf(ipaddress.ip_network)),
                "port_allowlist": config.Optional(config.TupleOf(int)),
                "port_denylist": config.Optional(config.TupleOf(int)),
                "hostname_denylist": config.Optional(config.TupleOf(config.String)),
                "allow_ipv6": config.Optional(config.Boolean, default=False),
            },
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    if options.pool_connections is not None:
        kwargs.setdefault("pool_connections", options.pool_connections)
    if options.pool_maxsize is not None:
        kwargs.setdefault("pool_maxsize", options.pool_maxsize)
    if options.max_retries is not None:
        kwargs.setdefault("max_retries", options.max_retries)
    if options.pool_block is not None:
        kwargs.setdefault("pool_block", options.pool_block)

    kwargs.setdefault(
        "validator",
        AddrValidator(
            ip_whitelist=options.filter.ip_allowlist,
            ip_blacklist=options.filter.ip_denylist,
            port_whitelist=options.filter.port_allowlist,
            port_blacklist=options.filter.port_denylist,
            hostname_blacklist=options.filter.hostname_denylist,
            allow_ipv6=options.filter.allow_ipv6,
        ),
    )
    return ValidatingHTTPAdapter(**kwargs)


PROM_NAMESPACE = "http_client"
HTTP_LABELS_COMMON = [
    "http_method",
    "http_client_name",
]
HTTP_LABELS_TERMINAL = [*HTTP_LABELS_COMMON, "http_success"]

# Latency histogram of HTTP calls made by clients
# buckets are defined above (from 100µs to ~14.9s)
LATENCY_SECONDS = Histogram(
    f"{PROM_NAMESPACE}_latency_seconds",
    "Latency histogram of HTTP calls made by clients",
    HTTP_LABELS_TERMINAL,
    buckets=default_latency_buckets,
)
# Counter counting total HTTP requests started by a given client
REQUESTS_TOTAL = Counter(
    f"{PROM_NAMESPACE}_requests_total",
    "Total number of HTTP requests started by a given client",
    [*HTTP_LABELS_TERMINAL, "http_response_code"],
)
# Gauge showing current number of active requests by a given client
ACTIVE_REQUESTS = Gauge(
    f"{PROM_NAMESPACE}_active_requests",
    "Number of active requests for a given client",
    HTTP_LABELS_COMMON,
)


class BaseplateSession:
    """A proxy for :py:class:`requests.Session`.

    Requests sent with this client will be instrumented automatically.

    """

    def __init__(
        self, adapter: HTTPAdapter, name: str, span: Span, client_name: Optional[str] = None
    ) -> None:
        self.adapter = adapter
        self.name = name
        self.span = span
        self.client_name = client_name

    def delete(self, url: str, **kwargs: Any) -> Response:
        """Send a DELETE request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("DELETE", url, **kwargs)

    def get(self, url: str, **kwargs: Any) -> Response:
        """Send a GET request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("GET", url, **kwargs)

    def head(self, url: str, **kwargs: Any) -> Response:
        """Send a HEAD request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("HEAD", url, **kwargs)

    def options(self, url: str, **kwargs: Any) -> Response:
        """Send an OPTIONS request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("OPTIONS", url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> Response:
        """Send a PATCH request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("PATCH", url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> Response:
        """Send a POST request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("POST", url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> Response:
        """Send a PUT request.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        return self.request("PUT", url, **kwargs)

    def prepare_request(self, request: Request) -> PreparedRequest:
        """Construct a :py:class:`~requests.PreparedRequest` for later use.

        The prepared request can be stored or manipulated and then used with
        :py:meth:`send`.

        """
        return request.prepare()

    def request(self, method: str, url: Union[str, bytes], **kwargs: Any) -> Response:
        """Send a request.

        :param method: The HTTP method of the request, e.g. ``GET``, ``PUT``, etc.
        :param url: The URL to send the request to.

        See :py:func:`requests.request` for valid keyword arguments.

        """
        send_kwargs = {
            "timeout": kwargs.pop("timeout", None),
            "allow_redirects": kwargs.pop("allow_redirects", None),
            "verify": kwargs.pop("verify", True),
        }
        request = Request(method=method.upper(), url=url, **kwargs)
        prepared = self.prepare_request(request)
        return self.send(prepared, **send_kwargs)

    def _add_span_context(self, span: Span, request: PreparedRequest) -> None:
        pass

    def send(self, request: PreparedRequest, **kwargs: Any) -> Response:
        """Send a :py:class:`~requests.PreparedRequest`."""
        active_request_label_values = {
            "http_method": request.method.lower() if request.method else "",
            "http_client_name": self.client_name if self.client_name is not None else self.name,
        }
        start_time = time.perf_counter()

        try:
            with self.span.make_child(f"{self.name}.request").with_tags(
                {
                    "http.url": request.url,
                    "http.method": request.method.lower() if request.method else "",
                    "http.slug": self.client_name if self.client_name is not None else self.name,
                }
            ) as span, ACTIVE_REQUESTS.labels(**active_request_label_values).track_inprogress():
                self._add_span_context(span, request)

                # we cannot re-use the same session every time because sessions re-use the same
                # CookieJar and so we'd muddle cookies cross-request. if the application wants
                # to keep track of cookies, it should do so itself.
                #
                # note: we're still getting connection pooling because we're re-using the adapter.
                session = Session()
                session.mount("http://", self.adapter)
                session.mount("https://", self.adapter)
                response = session.send(request, **kwargs)

                http_status_code = response.status_code
                span.set_tag("http.status_code", http_status_code)

            return response
        finally:
            if sys.exc_info()[0] is not None:
                status_code = ""
                http_success = "false"
            elif response and response.status_code:
                http_success = getHTTPSuccessLabel(response.status_code)
                status_code = str(response.status_code)
            else:
                status_code = ""
                http_success = ""

            latency_label_values = {**active_request_label_values, "http_success": http_success}
            requests_total_label_values = {
                **latency_label_values,
                "http_response_code": str(status_code),
            }

            LATENCY_SECONDS.labels(**latency_label_values).observe(time.perf_counter() - start_time)
            REQUESTS_TOTAL.labels(**requests_total_label_values).inc()


class InternalBaseplateSession(BaseplateSession):
    def _add_span_context(self, span: Span, request: PreparedRequest) -> None:
        request.headers["X-Trace"] = str(span.trace_id)
        request.headers["X-Parent"] = str(span.parent_id)
        request.headers["X-Span"] = str(span.id)
        if span.sampled:
            request.headers["X-Sampled"] = "1"
        if span.flags is not None:
            request.headers["X-Flags"] = str(span.flags)

        try:
            edge_context = span.context.raw_edge_context
        except AttributeError:
            pass
        else:
            request.headers["X-Edge-Request"] = base64.b64encode(edge_context).decode()


class RequestsContextFactory(ContextFactory):
    """Requests client context factory.

    This factory will attach a
    :py:class:`~baseplate.clients.requests.BaseplateSession` to an attribute
    on the :py:class:`~baseplate.RequestContext`. When HTTP requests are sent
    via this session, they will use connections from the provided
    :py:class:`~requests.adapters.HTTPAdapter` connection pools and
    automatically record diagnostic information.

    Note that though the connection pool is shared across calls, a new
    :py:class:`~requests.Session` is created for each request so that cookies
    and other state are not accidentally shared between requests. If you do
    want to persist state, you will need to do it in your application.

    :param adapter: A transport adapter for making HTTP requests. See
        :py:func:`http_adapter_from_config`.
    :param session_cls: The type for the actual session object to put on the
        request context.
    :param client_name: Custom name to be emitted under the http_client_name label
        for prometheus metrics. Defaults back to session_cls.name if None

    """

    def __init__(
        self,
        adapter: HTTPAdapter,
        session_cls: Type[BaseplateSession],
        client_name: Optional[str] = None,
    ) -> None:
        self.adapter = adapter
        self.session_cls = session_cls
        self.client_name = client_name

    def make_object_for_context(self, name: str, span: Span) -> BaseplateSession:
        return self.session_cls(self.adapter, name, span, client_name=self.client_name)


class InternalRequestsClient(config.Parser):
    """Configure a Requests client for use with internal Baseplate HTTP services.

    Requests made with this client **will** include trace context and
    :doc:`edge context </api/baseplate/lib/edgecontext>`. This client should
    only be used to speak to trusted internal services.  URLs that resolve to
    public addresses will be rejected.  It is not possible to override the
    Advocate address validator used by this client.

    .. warning:: Requesting user-specified URLs with this client could lead to
        `Server-Side Request Forgery`_. Ensure that you only request trusted URLs
        e.g. hard-coded or from a local configuration file.

    .. _`Server-Side Request Forgery`: https://en.wikipedia.org/wiki/Server-side_request_forgery

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`http_adapter_from_config` for available configuration settings.

    :param client_name: Custom name to be emitted under the http_client_name label
        for prometheus metrics. Defaults back to session_cls.name if None

    """

    def __init__(self, client_name: Optional[str] = None, **kwargs: Any) -> None:
        self.client_name = client_name
        self.kwargs = kwargs

        if "validator" in kwargs:
            raise Exception("validator is hard-coded for internal clients")

    def parse(self, key_path: str, raw_config: config.RawConfig) -> RequestsContextFactory:
        # use advocate to ensure this client only ever gets used
        # with internal services over the internal network.
        #
        # the allowlist takes precedence. allow loopback and private addresses,
        # deny the rest.
        validator = AddrValidator(
            ip_whitelist={
                ipaddress.ip_network("127.0.0.0/8"),
                ipaddress.ip_network("10.0.0.0/8"),
                ipaddress.ip_network("172.16.0.0/12"),
                ipaddress.ip_network("192.168.0.0/16"),
            },
            ip_blacklist={ipaddress.ip_network("0.0.0.0/0")},
            port_blacklist=[0],  # disable the default allowlist by giving an explicit denylist
            allow_ipv6=False,
        )

        adapter = http_adapter_from_config(
            raw_config, prefix=f"{key_path}.", validator=validator, **self.kwargs
        )
        return RequestsContextFactory(
            adapter, session_cls=InternalBaseplateSession, client_name=self.client_name
        )


class ExternalRequestsClient(config.Parser):
    """Configure a Requests client for use with external HTTP services.

    Requests made with this client **will not** include trace context and
    :doc:`edge context </api/baseplate/lib/edgecontext>`. This client is
    suitable for use with third party or untrusted services.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`http_adapter_from_config` for available configuration settings.

    :param client_name: Custom name to be emitted under the http_client_name label
        for prometheus metrics. Defaults back to session_cls.name if None

    """

    def __init__(self, client_name: Optional[str] = None, **kwargs: Any) -> None:
        self.client_name = client_name
        self.kwargs = kwargs

    def parse(self, key_path: str, raw_config: config.RawConfig) -> RequestsContextFactory:
        adapter = http_adapter_from_config(raw_config, f"{key_path}.", **self.kwargs)
        return RequestsContextFactory(
            adapter, session_cls=BaseplateSession, client_name=self.client_name
        )



from random import random
from typing import Any
from typing import Optional

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import LocalSpan
from baseplate import RequestContext
from baseplate import Span
from baseplate import SpanObserver
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.observers.timeout import ServerTimeout


class MetricsBaseplateObserver(BaseplateObserver):
    """Metrics collecting observer.

    This observer reports metrics to statsd. It does two important things:

    * it tracks the time taken in serving each request.
    * it batches all metrics generated during a request into as few packets
      as possible.

    The batch is accessible to your application during requests as the
    ``metrics`` attribute on the :py:class:`~baseplate.RequestContext`.

    :param client: The client where metrics will be sent.

    """

    def __init__(self, client: metrics.Client, sample_rate: float = 1.0):
        self.client = client
        self.sample_rate = sample_rate

    @classmethod
    def from_config_and_client(
        cls, raw_config: config.RawConfig, client: metrics.Client
    ) -> "MetricsBaseplateObserver":
        cfg = config.parse_config(
            raw_config,
            {"metrics_observer": {"sample_rate": config.Optional(config.Percent, default=1.0)}},
        )
        return cls(client, sample_rate=cfg.metrics_observer.sample_rate)

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        batch = self.client.batch()
        context.metrics = batch
        if self.sample_rate == 1.0 or random() < self.sample_rate:
            observer: SpanObserver = MetricsServerSpanObserver(batch, server_span, self.sample_rate)
        else:
            observer = MetricsServerSpanDummyObserver(batch)
        server_span.register(observer)


class MetricsServerSpanDummyObserver(SpanObserver):
    # for requests that aren't sampled
    def __init__(self, batch: metrics.Batch):
        self.batch = batch

    def on_start(self) -> None:
        pass

    def on_incr_tag(self, key: str, delta: float) -> None:
        pass

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.batch.flush()

    def on_child_span_created(self, span: Span) -> None:
        pass


class MetricsServerSpanObserver(SpanObserver):
    def __init__(self, batch: metrics.Batch, server_span: Span, sample_rate: float = 1.0):
        self.batch = batch
        self.base_name = "server." + server_span.name
        self.timer: Optional[metrics.Timer] = None
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.timer = self.batch.timer(self.base_name)
        self.timer.start(self.sample_rate)

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.batch.counter(key).increment(delta, sample_rate=self.sample_rate)

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        # the timer might not exist if another observer threw an exception
        # before we got our on_start() called
        if self.timer:
            self.timer.stop()

        if not exc_info:
            self.batch.counter(f"{self.base_name}.success").increment(sample_rate=self.sample_rate)
        else:
            self.batch.counter(f"{self.base_name}.failure").increment(sample_rate=self.sample_rate)

            if exc_info[0] is not None and issubclass(ServerTimeout, exc_info[0]):
                self.batch.counter(f"{self.base_name}.timed_out").increment(
                    sample_rate=self.sample_rate
                )

        self.batch.flush()

    def on_child_span_created(self, span: Span) -> None:
        observer: SpanObserver
        if isinstance(span, LocalSpan):
            observer = MetricsLocalSpanObserver(self.batch, span, self.sample_rate)
        else:
            observer = MetricsClientSpanObserver(self.batch, span, self.sample_rate)
        span.register(observer)


class MetricsLocalSpanObserver(SpanObserver):
    def __init__(self, batch: metrics.Batch, span: Span, sample_rate: float = 1.0):
        self.batch = batch
        self.timer = batch.timer(f"{span.component_name}.{span.name}")
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.timer.start(self.sample_rate)

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.batch.counter(key).increment(delta, sample_rate=self.sample_rate)

    def on_child_span_created(self, span: Span) -> None:
        observer: SpanObserver
        if isinstance(span, LocalSpan):
            observer = MetricsLocalSpanObserver(self.batch, span, self.sample_rate)
        else:
            observer = MetricsClientSpanObserver(self.batch, span, self.sample_rate)
        span.register(observer)

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.timer.stop()


class MetricsClientSpanObserver(SpanObserver):
    def __init__(self, batch: metrics.Batch, span: Span, sample_rate: float = 1.0):
        self.batch = batch
        self.base_name = f"clients.{span.name}"
        self.timer = batch.timer(self.base_name)
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.timer.start(self.sample_rate)

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.batch.counter(key).increment(delta, sample_rate=self.sample_rate)

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.timer.stop()
        suffix = "success" if not exc_info else "failure"
        self.batch.counter(self.base_name + "." + suffix).increment(sample_rate=self.sample_rate)

    def on_log(self, name: str, payload: Any) -> None:
        if name == "error.object":
            self.batch.counter(f"errors.{payload.__class__.__name__}").increment(
                sample_rate=self.sample_rate
            )



import unittest

from baseplate.lib import log_formatter


class CustomJSONFormatterTests(unittest.TestCase):
    def test_dict_with_correct_key(self):
        formatter = log_formatter.CustomJsonFormatter("")
        data = {"levelname": "foo"}
        assert formatter.process_log_record(data) == {"level": "foo"}

    def test_dict_without_correct_key(self):
        formatter = log_formatter.CustomJsonFormatter("")
        data = {"levelno": 1}
        assert formatter.process_log_record(data) == {"level": None, "levelno": 1}

    def test_wrong_type(self):
        formatter = log_formatter.CustomJsonFormatter("")
        with self.assertRaises(AttributeError):
            formatter.process_log_record("foo")



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Optional


class EdgeContextFactory(ABC):
    """Abstract base for a factory that parses edge context data."""

    @abstractmethod
    def from_upstream(self, header_value: Optional[bytes]) -> Any:
        """Parse a serialized edge context header from an inbound request.

        :param header_value: The raw bytes of the header payload.

        """



import time


class RateLimitBackend:
    """An interface for rate limit backends to implement."""

    def consume(self, key: str, amount: int, allowance: int, interval: int) -> bool:
        """Consume the given `amount` from the allowance for the given `key`.

        This will return true if the `key` remains below the `allowance`
        after consuming the given `amount`.

        :param key: The name of the rate limit bucket to consume from.
        :param amount: The amount to consume from the rate limit bucket.
        :param allowance: The maximum allowance for the rate limit bucket.
        :param interval: The interval to reset the allowance.

        """
        raise NotImplementedError


def _get_current_bucket(bucket_size: int) -> str:
    current_timestamp_seconds = int(time.time())
    return str(current_timestamp_seconds // bucket_size)



"""Helpers for interacting with ZooKeeper."""
from typing import Optional

from kazoo.client import KazooClient
from kazoo.handlers.gevent import SequentialGeventHandler

from baseplate.lib import config
from baseplate.lib.secrets import SecretsStore
from baseplate.server.monkey import gevent_is_patched


def zookeeper_client_from_config(
    secrets: SecretsStore, app_config: config.RawConfig, read_only: Optional[bool] = None
) -> KazooClient:
    """Configure and return a ZooKeeper client.

    There are several configuration options:

    ``zookeeper.hosts``
        A comma-delimited list of hosts with optional ``chroot`` at the end.
        For example ``zk01:2181,zk02:2181`` or
        ``zk01:2181,zk02:2181/some/root``.
    ``zookeeper.credentials``
        (Optional) A comma-delimited list of paths to secrets in the secrets
        store that contain ZooKeeper authentication credentials. Secrets should
        be of the "simple" type and contain ``username:password``.
    ``zookeeper.timeout``
        (Optional) A time span of how long to wait for each connection attempt.

    The client will attempt forever to reconnect on connection loss.

    :param secrets: A secrets store object
    :param raw_config: The application configuration which should have
        settings for the ZooKeeper client.
    :param read_only: Whether or not to allow connections to read-only
        ZooKeeper servers.

    """
    full_cfg = config.parse_config(
        app_config,
        {
            "zookeeper": {
                "hosts": config.String,
                "credentials": config.Optional(config.TupleOf(config.String), default=[]),
                "timeout": config.Optional(config.Timespan, default=config.Timespan("5 seconds")),
            }
        },
    )

    # pylint: disable=maybe-no-member
    cfg = full_cfg.zookeeper

    auth_data = []
    for path in cfg.credentials:
        credentials = secrets.get_simple(path)
        auth_data.append(("digest", credentials.decode("utf8")))

    # Kazoo needs to use a different handler with gevent.
    if gevent_is_patched():
        handler = SequentialGeventHandler()
    else:
        handler = None

    return KazooClient(
        cfg.hosts,
        timeout=cfg.timeout.total_seconds(),
        auth_data=auth_data,
        read_only=read_only,
        handler=handler,
        # this retry policy tells Kazoo how often it should attempt connections
        # to ZooKeeper from its worker thread/greenlet. when the connection is
        # lost during normal operation (i.e. after it was first established)
        # Kazoo will do retries quietly in the background while the application
        # continues forward. because of this, we want it to retry forever so
        # that it doesn't just give up at some point. the application can still
        # decide if it wants to exit after being disconnected for an amount of
        # time by polling the KazooClient.connected property.
        #
        # note: KazooClient.start() has a timeout parameter which defaults to
        # 15 seconds and controls the maximum amount of time start() will block
        # waiting for the background thread to confirm it has established a
        # connection. so even though we do infinite retries here, users of this
        # function can configure the amount of time they are willing to wait
        # for initial connection.
        connection_retry=dict(
            max_tries=-1,  # keep reconnecting forever
            delay=0.1,  # initial delay
            backoff=2,  # exponential backoff
            max_jitter=1,  # maximum amount to jitter sleeptimes
            max_delay=60,  # never wait longer than this
        ),
    )



"""Policies for retrying an operation safely."""
import time

from typing import Iterator
from typing import Optional


class RetryPolicy:
    """A policy for retrying operations.

    Policies are meant to be used as an iterable::

        for time_remaining in RetryPolicy.new(attempts=3):
            try:
                some_operation.do(timeout=time_remaining)
                break
            except SomeError:
                pass
        else:
            raise MaxRetriesError

    """

    def yield_attempts(self) -> Iterator[Optional[float]]:
        """Return an iterator which controls attempts.

        On each iteration, the iterator will yield the number of seconds left
        to retry, this should be used to set the timeout on the operation
        being carried out. If there is no maximum time remaining,
        :py:data:`None` is yielded instead.

        The iterable will raise :py:exc:`StopIteration` once the operation
        should not be retried any further.

        """
        raise NotImplementedError

    def __iter__(self) -> Iterator[Optional[float]]:
        """Return the result of :py:meth:`yield_attempts`.

        This allows policies to be directly iterated over.

        """
        return self.yield_attempts()

    @staticmethod
    def new(
        attempts: Optional[int] = None,
        budget: Optional[float] = None,
        backoff: Optional[float] = None,
    ) -> "RetryPolicy":
        """Create a new retry policy with the given constraints.

        :param attempts: The maximum number of times the operation can be
            attempted.
        :param budget: The maximum amount of time, in seconds, that the
            local service will wait for the operation to succeed.
        :param backoff: The base amount of time, in seconds, for
            exponential back-off between attempts. ``N`` in (``N *
            2**attempts``).

        """
        policy: RetryPolicy = IndefiniteRetryPolicy()

        if attempts is not None:
            policy = MaximumAttemptsRetryPolicy(policy, attempts)

        if budget is not None:
            policy = TimeBudgetRetryPolicy(policy, budget)

        if backoff is not None:
            policy = ExponentialBackoffRetryPolicy(policy, backoff)

        return policy


class IndefiniteRetryPolicy(RetryPolicy):  # pragma: noqa
    """Retry immediately forever."""

    def yield_attempts(self) -> Iterator[Optional[float]]:
        while True:
            yield None


class MaximumAttemptsRetryPolicy(RetryPolicy):
    """Constrain the total number of attempts."""

    def __init__(self, policy: RetryPolicy, attempts: int):
        self.subpolicy = policy
        self.attempts = attempts

    def yield_attempts(self) -> Iterator[Optional[float]]:
        for attempt_number, time_remaining in enumerate(self.subpolicy):
            if attempt_number == self.attempts:
                break
            yield time_remaining


class TimeBudgetRetryPolicy(RetryPolicy):
    """Constrain attempts to an overall time budget."""

    def __init__(self, policy: RetryPolicy, budget: float):
        assert budget >= 0, "The time budget must not be negative."
        self.subpolicy = policy
        self.budget = budget

    def yield_attempts(self) -> Iterator[Optional[float]]:
        start_time = time.time()

        yield self.budget

        for _ in self.subpolicy:
            elapsed = time.time() - start_time
            time_remaining = self.budget - elapsed
            if time_remaining <= 0:
                break
            yield time_remaining


class ExponentialBackoffRetryPolicy(RetryPolicy):
    """Sleep exponentially longer between attempts."""

    def __init__(self, policy: RetryPolicy, base: float):
        self.subpolicy = policy
        self.base = base

    def yield_attempts(self) -> Iterator[Optional[float]]:
        for attempt, time_remaining in enumerate(self.subpolicy):
            if attempt > 0:
                delay = self.base * 2.0 ** (attempt - 1.0)
                if time_remaining:
                    delay = min(delay, time_remaining)
                    time_remaining -= delay

                time.sleep(delay)

            yield time_remaining



import os
import socket
import unittest

from baseplate import BaseplateObserver
from baseplate import SpanObserver
from baseplate.lib.config import Endpoint
from baseplate.lib.edgecontext import EdgeContextFactory


def get_endpoint_or_skip_container(name, default_port):
    """Find a test server of the given type or raise SkipTest.

    This is useful for running tests in environments where we can't launch
    servers.

    If an environment variable like BASEPLATE_MEMCACHED_ADDR is present, that will
    override the default of {name}:{default_port}.

    """
    address = os.environ.get(f"BASEPLATE_{name.upper()}_ADDR", f"{name}:{default_port:d}")
    endpoint = Endpoint(address)

    try:
        sock = socket.socket(endpoint.family, socket.SOCK_STREAM)
        sock.settimeout(0.1)
        sock.connect(endpoint.address)
    except OSError:
        raise unittest.SkipTest("could not find %s server for integration tests" % name)
    else:
        sock.close()

    return endpoint


class TestSpanObserver(SpanObserver):
    def __init__(self, span):
        self.span = span
        self.on_start_called = False
        self.on_finish_called = False
        self.on_finish_exc_info = None
        self.tags = {}
        self.logs = []
        self.children = []

    def on_start(self):
        assert not self.on_start_called, "start was already called on this span"
        self.on_start_called = True

    def on_set_tag(self, key, value):
        self.tags[key] = value

    def assert_tag(self, key, value):
        assert key in self.tags, f"{key!r} not found in tags ({list(self.tags.keys())!r})"
        assert self.tags[key] == value, "tag {!r}: expected value {!r} but found {!r}".format(
            key, value, self.tags[key]
        )

    def on_log(self, name, payload):
        self.logs.append((name, payload))

    def on_finish(self, exc_info):
        assert not self.on_finish_called, "finish was already called on this span"
        self.on_finish_called = True
        self.on_finish_exc_info = exc_info

    def on_child_span_created(self, span):
        child = TestSpanObserver(span)
        self.children.append(child)
        span.register(child)

    def get_only_child(self):
        assert len(self.children) == 1, "observer has wrong number of children"
        return self.children[0]


class TestBaseplateObserver(BaseplateObserver):
    def __init__(self):
        self.children = []

    def get_only_child(self):
        assert len(self.children) == 1, "observer has wrong number of children"
        return self.children[0]

    def on_server_span_created(self, context, server_span):
        child = TestSpanObserver(server_span)
        self.children.append(child)
        server_span.register(child)


class FakeEdgeContextFactory(EdgeContextFactory):
    RAW_BYTES = b"raw_payload"
    DECODED_CONTEXT = "foo"

    def from_upstream(self, header_value):
        if header_value == self.RAW_BYTES:
            return self.DECODED_CONTEXT
        elif header_value is None:
            return None



from typing import Optional

import gevent

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate import SpanObserver
from baseplate.lib import config


# this deliberately inherits from BaseException rather than Exception, just
# like gevent.timeout.Timeout itself does, so that we don't get caught in
# "except Exception:" checks in the code being timed out.
class ServerTimeout(BaseException):
    def __init__(self, span_name: str, timeout_seconds: float, debug: bool):
        super().__init__()
        self.span_name = span_name
        self.timeout_seconds = timeout_seconds
        self.debug = debug


class TimeoutBaseplateObserver(BaseplateObserver):
    @classmethod
    def from_config(cls, app_config: config.RawConfig) -> "TimeoutBaseplateObserver":
        cfg = config.parse_config(
            app_config,
            {
                "server_timeout": {
                    "default": config.Optional(
                        config.TimespanOrInfinite,
                        default=config.InfiniteTimespan,
                    ),
                    "debug": config.Optional(config.Boolean, default=False),
                    "by_endpoint": config.DictOf(config.TimespanOrInfinite),
                }
            },
        )
        return cls(cfg.server_timeout)

    def __init__(self, timeout_config: config.ConfigNamespace):
        self.config = timeout_config

    def on_server_span_created(self, context: RequestContext, server_span: ServerSpan) -> None:
        timeout = self.config.by_endpoint.get(server_span.name, self.config.default)

        min_timeout = None
        if timeout is not config.InfiniteTimespan:
            min_timeout = timeout.total_seconds()

        try:
            deadline_budget = context.deadline_budget
            if deadline_budget:
                if not min_timeout or deadline_budget < min_timeout:
                    min_timeout = deadline_budget
        except AttributeError:
            # no deadline budget in request header
            pass

        if min_timeout:
            observer = TimeoutServerSpanObserver(server_span, min_timeout, self.config.debug)
            server_span.register(observer)


class TimeoutServerSpanObserver(SpanObserver):
    def __init__(self, span: ServerSpan, timeout_seconds: float, debug: bool):
        exception = ServerTimeout(span.name, timeout_seconds, debug)
        self.timeout = gevent.Timeout(timeout_seconds, exception)

    def on_start(self) -> None:
        self.timeout.start()

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.timeout.close()



"""Extensions to the standard library `random` module."""
import bisect
import random
import typing

from typing import Callable
from typing import Generic
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import TypeVar


T = TypeVar("T")


class WeightedLottery(Generic[T]):
    """A lottery where items can have different chances of selection.

    Items will be picked with chance proportional to their weight relative to
    the sum of all weights, so the higher the weight, the higher the chance of
    being picked.

    :param items: Items to choose from.
    :param weight_key: A function that takes an
        item in ``items`` and returns a non-negative integer
        weight for that item.
    :raises: :py:exc:`ValueError` if any weights are negative or there are no
        items.

    .. testsetup::

        import random
        from baseplate.lib.random import WeightedLottery
        random.seed(12345)

    An example of usage:

    .. doctest::

        >>> words = ["apple", "banana", "cantelope"]
        >>> lottery = WeightedLottery(words, weight_key=len)
        >>> lottery.pick()
        'banana'
        >>> lottery.sample(2)
        ['apple', 'cantelope']

    """

    def __init__(self, items: Iterable[T], weight_key: Callable[[T], int]):
        self.weights: List[int] = []
        self.items = list(items)
        if not self.items:
            raise ValueError("items must not be empty")

        # Build a running list of current total weight.  This allows us to do a
        # binary search for a random weight to get a choice quickly.
        accumulated_weight = 0
        for item in self.items:
            weight = weight_key(item)
            if weight < 0:
                raise ValueError(f"weight for {item!r} must be non-negative")
            accumulated_weight += weight
            self.weights.append(accumulated_weight)

        if accumulated_weight <= 0:
            raise ValueError("at least one item must have weight")

    def _pick_index(self) -> int:
        winning_ticket = random.random() * self.weights[-1]
        return bisect.bisect(self.weights, winning_ticket)

    def pick(self) -> T:
        """Pick a random element from the lottery."""
        winning_idx = self._pick_index()
        return self.items[winning_idx]

    def sample(self, sample_size: int) -> Iterable[T]:
        """Sample elements from the lottery without replacement.

        :param sample_size: The number of items to sample from the lottery.

        """
        if not 0 <= sample_size < len(self.items):
            raise ValueError("sample size is negative or larger than the population")

        already_picked: Set[int] = set()
        results: List[Optional[T]] = [None] * sample_size

        # we use indexes in the set so we don't add a hashability requirement
        # to the items in the population.
        for i in range(sample_size):
            picked_index = self._pick_index()
            while picked_index in already_picked:
                picked_index = self._pick_index()
            results[i] = self.items[picked_index]
            already_picked.add(picked_index)
        return typing.cast(List[T], results)
